# Copyright 2023 Acme Gating, LLC
# Copyright 2024 BMW Group
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import json
import logging
import queue
import threading
import time
from uuid import uuid4

from kazoo import exceptions as kze
from zuul.zk.vendor.states import (
    WatchedEvent,
)
from kazoo.protocol.states import (
    EventType,
    KazooState,
)

from zuul.zk.vendor.states import AddWatchMode
from zuul.zk.zkobject import ZKContext


class CacheSyncSentinel:
    def __init__(self):
        self.event = threading.Event()

    def set(self):
        return self.event.set()

    def wait(self, timeout=None):
        return self.event.wait(timeout)


class ZuulTreeCache(abc.ABC):
    log = logging.getLogger("zuul.zk.ZooKeeper")
    event_log = logging.getLogger("zuul.zk.cache.event")
    qsize_warning_threshold = 1024

    # Sentinel object that can be returned from preCacheHook to not
    # update the current object.
    STOP_OBJECT_UPDATE = object()

    def __init__(self, zk_client, root, async_worker=True):
        """Use watchers to keep a cache of local objects up to date.

        Typically this is used with async worker threads so that we
        release the kazoo callback resources as quickly as possible.
        This prevents large and active trees from starving other
        threads.

        In case the tree is small and rarely changed, async_worker may
        be set to False in order to avoid incurring the overhead of
        starting two extra threads.

        :param ZookeeperClient zk_client: The Zuul ZookeeperClient object
        :param str root: The root of the cache
        :param bool async_worker: Whether to start async worker threads

        """
        self.zk_client = zk_client
        self.client = zk_client.client
        self.root = root
        self.sync_path = f"{root}/_sync/{uuid4().hex}"
        self._last_event_warning = time.monotonic()
        self._last_playback_warning = time.monotonic()
        self._cached_objects = {}
        self._cached_paths = set()
        self._ready = threading.Event()
        self._init_lock = threading.Lock()
        self._stopped = False
        self._stop_workers = False
        self._event_queue = queue.Queue()
        self._playback_queue = queue.Queue()
        self._event_worker = None
        self._playback_worker = None
        self._async_worker = async_worker
        self._zk_context = ZKContext(self.zk_client, None, None, self.log)
        # The maximum zxid seen for any object creations or
        # modifications.  This is not updated on delete transactions.
        self._max_zxid = WatchedEvent.NO_ZXID

        self.client.add_listener(self._sessionListener)
        self._start()

    @property
    def max_zxid(self):
        # Until we're ready, the max_zxid is unreliable while we are
        # walking the tree to re-sync.
        if self._ready:
            return self._max_zxid
        return WatchedEvent.NO_ZXID

    def _bytesToDict(self, data):
        return json.loads(data.decode('utf8'))

    def _sessionListener(self, state):
        if state == KazooState.LOST:
            self._ready.clear()
            self._stop_workers = True
            self._event_queue.put(None)
            self._playback_queue.put(None)
        elif state == KazooState.CONNECTED and not self._stopped:
            self.client.handler.short_spawn(self._start)

    def _cacheListener(self, event):
        if self._async_worker:
            self._event_queue.put(event)
        else:
            self._handleCacheEvent(event)

    def _start(self):
        with self._init_lock:
            self.log.debug("Initialize cache at %s", self.root)

            self._ready.clear()
            self._stop_workers = True
            self._event_queue.put(None)
            self._playback_queue.put(None)

            # If we have an event worker (this is a re-init), then wait
            # for it to finish stopping.
            if self._event_worker:
                self._event_worker.join()
            # Replace the queue since any events from the previous
            # session aren't valid.
            old_event_queue = self._event_queue
            self._event_queue = queue.Queue()
            # Prepare (but don't start) the new worker.
            if self._async_worker:
                self._event_worker = threading.Thread(
                    target=self._eventWorker)
                self._event_worker.daemon = True

            if self._playback_worker:
                self._playback_worker.join()
            old_playback_queue = self._playback_queue
            self._playback_queue = queue.Queue()
            if self._async_worker:
                self._playback_worker = threading.Thread(
                    target=self._playbackWorker)
                self._playback_worker.daemon = True

            # Clear the stop flag and start the workers now that we
            # are sure that both have stopped and we have cleared the
            # queues.
            self._stop_workers = False
            if self._async_worker:
                self._event_worker.start()
                self._playback_worker.start()

            try:
                self.client.add_watch(
                    self.root, self._cacheListener,
                    AddWatchMode.PERSISTENT_RECURSIVE)
                self.client.ensure_path(self.root)
                self._walkTree()
                self._ready.set()
                # Transfer sentinel events
                for (old_queue, new_queue) in [
                        (old_event_queue, self._event_queue),
                        (old_playback_queue, self._playback_queue),
                ]:
                    while True:
                        try:
                            x = old_queue.get(block=False)
                            old_queue.task_done()
                            if isinstance(x, CacheSyncSentinel):
                                new_queue.put(x)
                        except queue.Empty:
                            break
                self.log.debug("Cache at %s is ready", self.root)
            except Exception:
                self.log.exception("Error initializing cache at %s", self.root)
                self.client.handler.short_spawn(self._start)

    def stop(self):
        self._stopped = True
        self._event_queue.put(None)
        self._playback_queue.put(None)
        self.client.remove_listener(self._sessionListener)

    def _walkTree(self, root=None, seen_paths=None):
        # Recursively walk the tree and emit fake changed events for
        # every item in zk and fake deleted events for every item in
        # the cache that is not in zk
        exists = True
        am_root = False
        if root is None:
            am_root = True
            root = self.root
            seen_paths = set()
        if exists:
            seen_paths.add(root)
            event = WatchedEvent(EventType.NONE,
                                 self.client._state,
                                 root, WatchedEvent.NO_ZXID)
            self._cacheListener(event)
            try:
                for child in self.client.get_children(root):
                    safe_root = root
                    if safe_root == '/':
                        safe_root = ''
                    new_path = '/'.join([safe_root, child])
                    self._walkTree(new_path, seen_paths)
            except kze.NoNodeError:
                self.log.debug("Can't sync non-existent node %s", root)
        if am_root:
            for path in self._cached_paths.copy():
                if path not in seen_paths:
                    event = WatchedEvent(
                        EventType.NONE,
                        self.client._state,
                        path, WatchedEvent.NO_ZXID)
                    self._cacheListener(event)

    def _eventWorker(self):
        while not (self._stopped or self._stop_workers):
            event = self._event_queue.get()
            if event is None:
                self._event_queue.task_done()
                continue
            if isinstance(event, CacheSyncSentinel):
                event.set()
                self._event_queue.task_done()
                continue

            qsize = self._event_queue.qsize()
            if qsize > self.qsize_warning_threshold:
                now = time.monotonic()
                if now - self._last_event_warning > 60:
                    self.log.warning("Event queue size for cache at %s is %s",
                                     self.root, qsize)
                    self._last_event_warning = now

            try:
                self._handleCacheEvent(event)
            except Exception:
                self.log.exception("Error handling event %s:", event)
            self._event_queue.task_done()

    def _handleCacheEvent(self, event):
        # Ignore root node since we don't maintain a cached object for
        # it (all cached objects are under the root in our tree
        # caches).
        if event.path == self.root:
            return

        # Start by assuming we need to fetch data for the event.
        fetch = True
        if event.type == EventType.NONE:
            if event.path is None:
                # We're probably being told of a connection change; ignore.
                return
        elif (event.type == EventType.DELETED):
            # If this is a normal deleted event, we don't need to
            # fetch anything.
            fetch = False

        key, should_fetch = self.parsePath(event.path)

        if ((not should_fetch) and event.type != EventType.NONE
                and event.path != self.sync_path):
            # The cache doesn't care about this path, so we don't need
            # to fetch (unless the type is none (re-initialization) in
            # which case we always need to fetch in order to determine
            # existence).
            fetch = False

        if fetch:
            future = self.client.get_async(event.path)
        else:
            future = None
        if self._async_worker:
            self._playback_queue.put((event, future, key))
        else:
            self._handlePlayback(event, future, key)

    def _playbackWorker(self):
        while not (self._stopped or self._stop_workers):
            item = self._playback_queue.get()
            if item is None:
                self._playback_queue.task_done()
                continue
            if isinstance(item, CacheSyncSentinel):
                item.set()
                self._playback_queue.task_done()
                continue

            qsize = self._playback_queue.qsize()
            if qsize > self.qsize_warning_threshold:
                now = time.monotonic()
                if now - self._last_playback_warning > 60:
                    self.log.warning(
                        "Playback queue size for cache at %s is %s",
                        self.root, qsize)
                    self._last_playback_warning = now

            event, future, key = item
            try:
                self._handlePlayback(event, future, key)
            except Exception:
                self.log.exception("Error playing back event %s:", event)
            self._playback_queue.task_done()

    def _handlePlayback(self, event, future, key):
        # self.event_log.debug("Cache playback event %s", event)
        exists = None
        data, stat = None, None
        zxid = WatchedEvent.NO_ZXID

        if future:
            try:
                data, stat = future.get()
                zxid = stat.mzxid
                exists = True
            except kze.NoNodeError:
                exists = False

        # We set "exists" above in case of cache re-initialization,
        # which happens out of sequence with the normal watch events.
        # and we can't be sure whether the node still exists or not by
        # the time we process it.  Later cache watch events may
        # supercede this (for example, we may process a NONE event
        # here which we interpret as a delete which may be followed by
        # a normal delete event.  That case, and any other variations
        # should be anticipated.

        # If the event tells us whether the node exists, prefer that
        # value, otherwise fallback to what we determined above.
        if (event.type in (EventType.CREATED, EventType.CHANGED)):
            # Update our max_zxid only for created or changed events.
            # We receive zxids for deleted events, but we won't see
            # them on re-initialization.
            if event.zxid != WatchedEvent.NO_ZXID:
                # Prefer the zxid from the event, only use the zxid
                # from the stat call above if we don't have an event
                # (i.e., recovery).
                zxid = event.zxid
            exists = True
        elif (event.type == EventType.DELETED):
            exists = False

        self._max_zxid = max(zxid, self._max_zxid)

        # Keep the cached paths up to date
        if exists:
            self._cached_paths.add(event.path)
        else:
            self._cached_paths.discard(event.path)

        # Some caches have special handling for certain sub-objects
        if (self.preCacheHook(event, exists, data, stat)
                == self.STOP_OBJECT_UPDATE):
            return

        # If we don't actually cache this kind of object, return now
        if key is None:
            return

        obj = None
        if data:
            # Perform an in-place update of the cached object if possible
            obj = self._cached_objects.get(key)
            if obj:
                # Don't update to older data
                # Don't update a locked object
                if (stat.mzxid > obj._zstat.mzxid and
                    getattr(obj, 'lock', None) is None):
                    self.updateFromRaw(obj, key, data, stat)
            else:
                obj = self.objectFromRaw(key, data, stat)
                self._cached_objects[key] = obj
        else:
            try:
                obj = self._cached_objects[key]
                del self._cached_objects[key]
            except KeyError:
                # If it's already gone, don't care
                pass
        self.postCacheHook(event, data, stat, key, obj)

    def ensureReady(self):
        self._ready.wait()

    def waitForSync(self, timeout=None):
        # Wait for (at least) all the events that were received at the
        # time of this call to have been processed.
        if not self._async_worker:
            return True

        event_sentinel = CacheSyncSentinel()
        self._event_queue.put(event_sentinel)
        start = time.monotonic()
        if event_sentinel.wait(timeout):
            if timeout is not None:
                elapsed = time.monotonic() - start
                remain = timeout - elapsed
                if remain <= 0:
                    return False
                timeout = remain
            play_sentinel = CacheSyncSentinel()
            self._playback_queue.put(play_sentinel)
            return play_sentinel.wait(timeout)
        return False

    # Methods for subclasses:
    def preCacheHook(self, event, exists, data=None, stat=None):
        """Called before the cache is updated

        This is called for any add/update/remove event under the root,
        even for paths that are ignored, so users much test the
        relevance of the path in this method.

        The ``exists`` argument is provided in all cases. In the case
        of EventType.NONE events, it indicates whether the cache has
        seen the node in ZK immediately before calling this method.
        Otherwise, it indicates whether or not the EventType would
        cause the node to exist in ZK.

        Return sentinel STOP_OBJECT_UPDATE to skip the object update
        step.

        :param EventType event: The event.
        :param bool exists: Whether the object exists in ZK.
        :param bytes data: data when fetched, else None
        :param ZnodeStat stat: ZNode stat when the node exists, else None

        """
        return None

    def postCacheHook(self, event, data, stat, key, obj):
        """Called after the cache has been updated"""
        return None

    @abc.abstractmethod
    def parsePath(self, path):
        """Parse the path and return a cache key

        The cache key is an opaque object ignored by the cache, but
        must be hashable.

        A convention is to use a tuple of relevant path components as
        the key.

        Returns a tuple: (key, should_fetch)

        Return a key of None to indicate the path is not relevant to the cache.
        The should_fetch return value indicates whether the cache should
          fetch the contents of the path.
        """
        return None

    @abc.abstractmethod
    def objectFromRaw(self, key, data, stat):
        """Construct an object from ZooKeeper data

        Given data from ZK and cache key, construct
        and return an object to insert into the cache.

        :param object key: The key as returned by parsePath.
        :param dict data: The raw data.
        :param Zstat stat: The zstat of the znode.
        """
        pass

    @abc.abstractmethod
    def updateFromRaw(self, obj, key, data, stat):
        """Construct an object from ZooKeeper data

        Update an existing object with new data.

        :param object obj: The old object.
        :param object key: The key as returned by parsePath.
        :param dict data: The raw data.
        :param Zstat stat: The zstat of the znode.
        """
        pass



# Copyright 2012 Hewlett-Packard Development Company, L.P.
# Copyright 2013 OpenStack Foundation
# Copyright 2023 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import voluptuous as v

from zuul.trigger import BaseTrigger
from zuul.driver.timer.timermodel import TimerEventFilter
from zuul.driver.util import to_list, make_regex


class TimerTrigger(BaseTrigger):
    name = 'timer'

    def getEventFilters(self, connection_name, trigger_conf,
                        parse_context):
        efilters = []
        for trigger in to_list(trigger_conf):
            types = [make_regex('timer')]
            f = TimerEventFilter(
                connection_name=connection_name,
                trigger=self,
                types=types,
                timespecs=to_list(trigger['time']),
                dereference=trigger.get('dereference', False),
                debug=trigger.get('debug'),
            )

            efilters.append(f)

        return efilters


def getSchema():
    timer_trigger = {
        v.Required('time'): str,
        'dereference': bool,
        'debug': bool,
    }
    return timer_trigger



# Copyright 2014 Rackspace Australia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import smtplib

from email.mime.text import MIMEText

from zuul.connection import BaseConnection
from zuul.lib.logutil import get_annotated_logger


class SMTPConnection(BaseConnection):
    driver_name = 'smtp'
    log = logging.getLogger("zuul.SMTPConnection")

    def __init__(self, driver, connection_name, connection_config):
        super(SMTPConnection, self).__init__(driver, connection_name,
                                             connection_config)

        self.smtp_server = self.connection_config.get(
            'server', 'localhost')
        self.smtp_port = self.connection_config.get('port', 25)
        self.smtp_default_from = self.connection_config.get(
            'default_from', 'zuul')
        self.smtp_default_to = self.connection_config.get(
            'default_to', 'zuul')
        self.smtp_user = self.connection_config.get('user')
        self.smtp_pass = self.connection_config.get('password')
        starttls = self.connection_config.get('use_starttls', 'false')
        if starttls.lower() == 'false':
            self.smtp_starttls = False
        else:
            self.smtp_starttls = True

    def sendMail(self, subject, message, from_email=None, to_email=None,
                 zuul_event_id=None):
        # Create a text/plain email message
        from_email = from_email \
            if from_email is not None else self.smtp_default_from
        to_email = to_email if to_email is not None else self.smtp_default_to

        msg = MIMEText(message)
        msg['Subject'] = subject
        msg['From'] = from_email
        msg['To'] = to_email

        try:
            s = smtplib.SMTP(self.smtp_server, self.smtp_port)
            if self.smtp_starttls:
                s.starttls()
                s.ehlo()
            if self.smtp_user is not None and self.smtp_pass is not None:
                s.login(self.smtp_user, self.smtp_pass)
            s.sendmail(from_email, to_email.split(','), msg.as_string())
            s.quit()
        except Exception as e:
            log = get_annotated_logger(self.log, zuul_event_id)
            log.warning("Error sending mail via SMTP: %s", e)



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io
import json
import logging
import os
import re
import textwrap
import yaml
from unittest import skip
from datetime import datetime, timedelta

from tests.base import AnsibleZuulTestCase, zuul_config


class FunctionalZuulStreamMixIn:
    tenant_config_file = 'config/remote-zuul-stream/main.yaml'
    # This should be overriden in child classes.
    ansible_version = 'X'
    ansible_core_version = 'X.Y'

    def _setUp(self):
        self.log_console_port = 19000 + int(
            self.ansible_core_version.split('.')[1])
        self.executor_server.log_console_port = self.log_console_port
        self.wait_timeout = 180
        self.fake_nodepool.remote_ansible = True
        # This catches the Ansible output; rather than the callback
        # output captured in the job log.  For example if the callback
        # fails, there will be an error output in this stream.
        self.logger = logging.getLogger('zuul.AnsibleJob')
        self.console_output = io.StringIO()
        self.logger.addHandler(logging.StreamHandler(self.console_output))

        ansible_remote = os.environ.get('ZUUL_REMOTE_IPV4')
        self.assertIsNotNone(ansible_remote)

    def _run_job(self, job_name, create=True, nodes=None, split=False):
        # Keep the jobdir around so we can inspect contents if an
        # assert fails. It will be cleaned up anyway as it is contained
        # in a tmp dir which gets cleaned up after the test.
        self.executor_server.keep_jobdir = True

        if nodes is None:
            nodes = [
                {'name': 'compute1',
                 'label': 'whatever'},
                {'name': 'controller',
                 'label': 'whatever'},
            ]

        # Output extra ansible info so we might see errors.
        self.executor_server.verbose = True
        if create:
            conf = [
                {
                    'job': {
                        'name': job_name,
                        'run': f'playbooks/{job_name}.yaml',
                        'ansible-version': self.ansible_version,
                        'ansible-split-streams': split,
                        'vars': {
                            'test_console_port': self.log_console_port,
                        },
                        'roles': [
                            {'zuul': 'org/common-config'},
                        ],
                        'nodeset': {'nodes': nodes},
                    }
                }, {
                    'project': {'check': {'jobs': [job_name]}}
                }
            ]
            conf = yaml.safe_dump(conf)
        else:
            conf = textwrap.dedent(
                """
                - project:
                    check:
                      jobs:
                        - {job_name}
                """.format(job_name=job_name))
        file_dict = {'zuul.yaml': conf}
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A',
                                           files=file_dict)
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        job = self.getJobFromHistory(job_name)
        return job

    def _get_job_output(self, build):
        path = os.path.join(self.jobdir_root, build.uuid,
                            'work', 'logs', 'job-output.txt')
        with open(path) as f:
            return f.read()

    def _get_job_json(self, build):
        path = os.path.join(self.jobdir_root, build.uuid,
                            'work', 'logs', 'job-output.json')
        with open(path) as f:
            return json.loads(f.read())

    def _assertLogLine(self, line, log, full_match=True):
        pattern = (r'^\d\d\d\d-\d\d-\d\d \d\d:\d\d\:\d\d\.\d\d\d\d\d\d \| %s%s'
                   % (line, '$' if full_match else ''))
        log_re = re.compile(pattern, re.MULTILINE)
        m = log_re.search(log)
        if m is None:
            raise Exception("'%s' not found in log" % (line,))

    def assertLogLineStartsWith(self, line, log):
        self._assertLogLine(line, log, full_match=False)

    def assertLogLine(self, line, log):
        self._assertLogLine(line, log, full_match=True)

    def _getLogTime(self, line, log):
        pattern = (r'^(\d\d\d\d-\d\d-\d\d \d\d:\d\d\:\d\d\.\d\d\d\d\d\d)'
                   r' \| %s\n'
                   r'(\d\d\d\d-\d\d-\d\d \d\d:\d\d\:\d\d\.\d\d\d\d\d\d)'
                   % line)
        log_re = re.compile(pattern, re.MULTILINE)
        m = log_re.search(log)
        if m is None:
            raise Exception("'%s' not found in log" % (line,))
        else:
            date1 = datetime.strptime(m.group(1), "%Y-%m-%d %H:%M:%S.%f")
            date2 = datetime.strptime(m.group(2), "%Y-%m-%d %H:%M:%S.%f")
            return (date1, date2)

    def test_command(self):
        job = self._run_job('command')
        with self.jobLog(job):
            build = self.history[-1]
            self.assertEqual(build.result, 'SUCCESS')

            console_output = self.console_output.getvalue()
            # This should be generic enough to match any callback
            # plugin failures, which look something like
            #
            #  [WARNING]: Failure using method (v2_runner_on_ok) in \
            #                                                  callback plugin
            #  (<ansible.plugins.callback.zuul_stream.CallbackModule object at'
            #  0x7f89f72a20b0>): 'dict' object has no attribute 'startswith'"
            #  Callback Exception:
            #  ...
            #
            self.assertNotIn('[WARNING]: Failure using method', console_output)

            text = self._get_job_output(build)
            data = self._get_job_json(build)

            token_stdout = "Standard output test {}".format(
                self.history[0].jobdir.src_root)
            token_stderr = "Standard error test {}".format(
                self.history[0].jobdir.src_root)
            result = data[0]['plays'][1]['tasks'][2]['hosts']['compute1']
            self.assertEqual("\n".join((token_stdout, token_stderr)),
                             result['stdout'])
            self.assertEqual("", result['stderr'])

            # Find the "creates" tasks
            create1_task = data[0]['plays'][4]['tasks'][3]
            create1_host = create1_task['hosts']['compute1']
            self.assertIsNotNone(create1_host['delta'])
            self.assertNotIn("Did not run command since", create1_host['msg'])
            self.assertEqual("Creates file that does not exist",
                             create1_task['task']['name'])
            create2_task = data[0]['plays'][4]['tasks'][4]
            create2_host = create2_task['hosts']['compute1']
            self.assertIsNone(create2_host['delta'])
            self.assertIn("Did not run command since", create2_host['msg'])
            self.assertEqual("Creates file that already exists",
                             create2_task['task']['name'])
            self.assertLogLine(r'compute1 \| ok: Runtime: None', text)

            # Test no_log for command
            nolog_task = data[0]['plays'][1]['tasks'][15]
            self.assertEqual("Test no_log command",
                             nolog_task['task']['name'])
            nolog_json_as_str = json.dumps(nolog_task)
            self.assertNotIn("This task should not log", nolog_json_as_str)
            self.assertNotIn('This task should not log', text)

            # Test no_log for shell
            nolog_task = data[0]['plays'][1]['tasks'][16]
            self.assertEqual("Test no_log shell",
                             nolog_task['task']['name'])
            nolog_json_as_str = json.dumps(nolog_task)
            self.assertNotIn("This task should not log", nolog_json_as_str)
            self.assertNotIn('This task should not log', text)

            # Test it does not appear in the registered output from command
            # NOTE: This isn't necessarily a desired behavior, but it is
            # a side effect of how we prevent output in the JSON log; we
            # could consider changing it, but we include assertions so that
            # we know that the behavior is consistent over time.
            nolog_task = data[0]['plays'][1]['tasks'][17]
            self.assertEqual("Inspect command_no_log_result",
                             nolog_task['task']['name'])
            nolog_json_as_str = json.dumps(nolog_task)
            self.assertNotIn("This task should not log", nolog_json_as_str)

            # Test it does not appear in the registered output from shell
            nolog_task = data[0]['plays'][1]['tasks'][18]
            self.assertEqual("Inspect shell_no_log_result",
                             nolog_task['task']['name'])
            nolog_json_as_str = json.dumps(nolog_task)
            self.assertNotIn("This task should not log", nolog_json_as_str)

            self.assertLogLine(
                r'RUN START: \[untrusted : review.example.com/org/project/'
                r'playbooks/command.yaml@master\]', text)
            self.assertLogLine(r'PLAY \[all\]', text)
            self.assertLogLine(
                r'Ansible version={}'.format(self.ansible_core_version), text)
            self.assertLogLine(r'TASK \[Show contents of first file\]', text)
            self.assertLogLine(r'controller \| command test one', text)
            self.assertLogLine(
                r'controller \| ok: Runtime: \d:\d\d:\d\d\.\d\d\d\d\d\d', text)
            self.assertLogLine(r'TASK \[Show contents of second file\]', text)
            self.assertLogLine(r'compute1 \| command test two', text)
            self.assertLogLine(r'controller \| command test two', text)
            self.assertLogLine(r'compute1 \| This is a rescue task', text)
            self.assertLogLine(r'controller \| This is a rescue task', text)
            self.assertLogLine(r'compute1 \| This is an always task', text)
            self.assertLogLine(r'controller \| This is an always task', text)
            for i in range(2):
                self.assertLogLine(r'compute1 \| This is a command task from '
                                   r'include_tasks inside a loop '
                                   rf'\(index={i}\)',
                                   text)
                self.assertLogLine(r'controller \| This is a shell task from '
                                   r'include_tasks inside a loop '
                                   rf'\(index={i}\)',
                                   text)
            self.assertLogLine(r'compute1 \| This is a handler', text)
            self.assertLogLine(r'controller \| This is a handler', text)
            self.assertLogLine(r'controller \| First free task', text)
            self.assertLogLine(r'controller \| Second free task', text)
            self.assertLogLine(r'controller \| This is a shell task after an '
                               'included role', text)
            self.assertLogLine(r'compute1 \| This is a shell task after an '
                               'included role', text)
            self.assertLogLine(r'controller \| This is a command task after '
                               'an included role', text)
            self.assertLogLine(r'compute1 \| This is a command task after an '
                               'included role', text)
            self.assertLogLine(r'controller \| This is a shell task with '
                               'delegate compute1', text)
            self.assertLogLine(r'controller \| This is a shell task with '
                               'delegate controller', text)
            self.assertLogLine(r'compute1 \| item_in_loop1', text)
            self.assertLogLine(r'compute1 \| ok: Item: item_in_loop1 '
                               r'Runtime: \d:\d\d:\d\d\.\d\d\d\d\d\d', text)
            self.assertLogLine(r'compute1 \| item_in_loop2', text)
            self.assertLogLine(r'compute1 \| ok: Item: item_in_loop2 '
                               r'Runtime: \d:\d\d:\d\d\.\d\d\d\d\d\d', text)
            self.assertLogLine(r'compute1 \| failed_in_loop1', text)
            self.assertLogLine(r'compute1 \| ok: Item: failed_in_loop1 '
                               r'Result: 1', text)
            self.assertLogLine(r'compute1 \| failed_in_loop2', text)
            self.assertLogLine(r'compute1 \| ok: Item: failed_in_loop2 '
                               r'Result: 1', text)
            self.assertLogLine(r'compute1 \| transitive-one', text)
            self.assertLogLine(r'compute1 \| transitive-two', text)
            self.assertLogLine(r'compute1 \| transitive-three', text)
            self.assertLogLine(r'compute1 \| transitive-four', text)
            self.assertLogLine(
                r'controller \| ok: Runtime: \d:\d\d:\d\d\.\d\d\d\d\d\d', text)
            self.assertLogLine('PLAY RECAP', text)
            self.assertLogLine(
                r'controller \| ok: \d+ changed: \d+ unreachable: 0 failed: 0 '
                'skipped: 2 rescued: 1 ignored: 0', text)
            self.assertLogLine(
                r'RUN END RESULT_NORMAL: \[untrusted : review.example.com/'
                r'org/project/playbooks/command.yaml@master]', text)
            time1, time2 = self._getLogTime(r'TASK \[Command Not Found\]',
                                            text)
            self.assertLess((time2 - time1) / timedelta(milliseconds=1),
                            9000)

            # This is from the debug: msg='{{ ansible_version }}'
            # testing raw variable output.  To make it version
            # agnostic, match just the start of
            #  compute1 | ok: {'string': '2.9.27'...

            # NOTE(ianw) 2022-08-24 : I don't know why the callback
            # for debug: msg= doesn't put the hostname first like
            # other output. Undetermined if bug or feature.
            self.assertLogLineStartsWith(
                r"""\{'string': '\d.""", text)
            # ... handling loops is a different path, and that does
            self.assertLogLineStartsWith(
                r"""compute1 \| ok: \{'string': '\d.""", text)
            self.assertLogLine(
                r'fake \| skipping: Conditional result was False', text)
            self.assertLogLine(r'compute1 \| Testing raw', text)

    def test_command_split_streams(self):
        job = self._run_job('command', split=True)
        with self.jobLog(job):
            build = self.history[-1]
            self.assertEqual(build.result, 'SUCCESS')

            console_output = self.console_output.getvalue()
            # This should be generic enough to match any callback
            # plugin failures, which look something like
            #
            #  [WARNING]: Failure using method (v2_runner_on_ok) in \
            #                                                  callback plugin
            #  (<ansible.plugins.callback.zuul_stream.CallbackModule object at'
            #  0x7f89f72a20b0>): 'dict' object has no attribute 'startswith'"
            #  Callback Exception:
            #  ...
            #
            self.assertNotIn('[WARNING]: Failure using method', console_output)

            text = self._get_job_output(build)
            data = self._get_job_json(build)

            token_stdout = "Standard output test {}".format(
                self.history[0].jobdir.src_root)
            token_stderr = "Standard error test {}".format(
                self.history[0].jobdir.src_root)
            result = data[0]['plays'][1]['tasks'][2]['hosts']['compute1']
            self.assertEqual(token_stdout, result['stdout'])
            self.assertEqual(token_stderr, result['stderr'])

            # Find the "creates" tasks
            create1_task = data[0]['plays'][4]['tasks'][3]
            create1_host = create1_task['hosts']['compute1']
            self.assertIsNotNone(create1_host['delta'])
            self.assertNotIn("Did not run command since", create1_host['msg'])
            self.assertEqual("Creates file that does not exist",
                             create1_task['task']['name'])
            create2_task = data[0]['plays'][4]['tasks'][4]
            create2_host = create2_task['hosts']['compute1']
            self.assertIsNone(create2_host['delta'])
            self.assertIn("Did not run command since", create2_host['msg'])
            self.assertEqual("Creates file that already exists",
                             create2_task['task']['name'])
            self.assertLogLine(r'compute1 \| ok: Runtime: None', text)

            # Test no_log for command
            nolog_task = data[0]['plays'][1]['tasks'][15]
            self.assertEqual("Test no_log command",
                             nolog_task['task']['name'])
            nolog_json_as_str = json.dumps(nolog_task)
            self.assertNotIn("This task should not log", nolog_json_as_str)
            self.assertNotIn('This task should not log', text)

            # Test no_log for shell
            nolog_task = data[0]['plays'][1]['tasks'][16]
            self.assertEqual("Test no_log shell",
                             nolog_task['task']['name'])
            nolog_json_as_str = json.dumps(nolog_task)
            self.assertNotIn("This task should not log", nolog_json_as_str)
            self.assertNotIn('This task should not log', text)

            # Test it does not appear in the registered output from command
            nolog_task = data[0]['plays'][1]['tasks'][17]
            self.assertEqual("Inspect command_no_log_result",
                             nolog_task['task']['name'])
            nolog_json_as_str = json.dumps(nolog_task)
            self.assertNotIn("This task should not log", nolog_json_as_str)

            # Test it does not appear in the registered output from shell
            nolog_task = data[0]['plays'][1]['tasks'][18]
            self.assertEqual("Inspect shell_no_log_result",
                             nolog_task['task']['name'])
            nolog_json_as_str = json.dumps(nolog_task)
            self.assertNotIn("This task should not log", nolog_json_as_str)

            self.assertLogLine(
                r'RUN START: \[untrusted : review.example.com/org/project/'
                r'playbooks/command.yaml@master\]', text)
            self.assertLogLine(r'PLAY \[all\]', text)
            self.assertLogLine(
                r'Ansible version={}'.format(self.ansible_core_version), text)
            self.assertLogLine(r'TASK \[Show contents of first file\]', text)
            self.assertLogLine(r'controller \| command test one', text)
            self.assertLogLine(
                r'controller \| ok: Runtime: \d:\d\d:\d\d\.\d\d\d\d\d\d', text)
            self.assertLogLine(r'TASK \[Show contents of second file\]', text)
            self.assertLogLine(r'compute1 \| command test two', text)
            self.assertLogLine(r'controller \| command test two', text)
            self.assertLogLine(r'compute1 \| This is a rescue task', text)
            self.assertLogLine(r'controller \| This is a rescue task', text)
            self.assertLogLine(r'compute1 \| This is an always task', text)
            self.assertLogLine(r'controller \| This is an always task', text)
            self.assertLogLine(r'compute1 \| This is a handler', text)
            self.assertLogLine(r'controller \| This is a handler', text)
            self.assertLogLine(r'controller \| First free task', text)
            self.assertLogLine(r'controller \| Second free task', text)
            self.assertLogLine(r'controller \| This is a shell task after an '
                               'included role', text)
            self.assertLogLine(r'compute1 \| This is a shell task after an '
                               'included role', text)
            self.assertLogLine(r'controller \| This is a command task after '
                               'an included role', text)
            self.assertLogLine(r'compute1 \| This is a command task after an '
                               'included role', text)
            self.assertLogLine(r'controller \| This is a shell task with '
                               'delegate compute1', text)
            self.assertLogLine(r'controller \| This is a shell task with '
                               'delegate controller', text)
            self.assertLogLine(r'compute1 \| item_in_loop1', text)
            self.assertLogLine(r'compute1 \| ok: Item: item_in_loop1 '
                               r'Runtime: \d:\d\d:\d\d\.\d\d\d\d\d\d', text)
            self.assertLogLine(r'compute1 \| item_in_loop2', text)
            self.assertLogLine(r'compute1 \| ok: Item: item_in_loop2 '
                               r'Runtime: \d:\d\d:\d\d\.\d\d\d\d\d\d', text)
            self.assertLogLine(r'compute1 \| failed_in_loop1', text)
            self.assertLogLine(r'compute1 \| ok: Item: failed_in_loop1 '
                               r'Result: 1', text)
            self.assertLogLine(r'compute1 \| failed_in_loop2', text)
            self.assertLogLine(r'compute1 \| ok: Item: failed_in_loop2 '
                               r'Result: 1', text)
            self.assertLogLine(r'compute1 \| transitive-one', text)
            self.assertLogLine(r'compute1 \| transitive-two', text)
            self.assertLogLine(r'compute1 \| transitive-three', text)
            self.assertLogLine(r'compute1 \| transitive-four', text)
            self.assertLogLine(
                r'controller \| ok: Runtime: \d:\d\d:\d\d\.\d\d\d\d\d\d', text)
            self.assertLogLine('PLAY RECAP', text)
            self.assertLogLine(
                r'controller \| ok: \d+ changed: \d+ unreachable: 0 failed: 0 '
                'skipped: 2 rescued: 1 ignored: 0', text)
            self.assertLogLine(
                r'RUN END RESULT_NORMAL: \[untrusted : review.example.com/'
                r'org/project/playbooks/command.yaml@master]', text)
            time1, time2 = self._getLogTime(r'TASK \[Command Not Found\]',
                                            text)
            self.assertLess((time2 - time1) / timedelta(milliseconds=1),
                            9000)

            # This is from the debug: msg='{{ ansible_version }}'
            # testing raw variable output.  To make it version
            # agnostic, match just the start of
            #  compute1 | ok: {'string': '2.9.27'...

            # NOTE(ianw) 2022-08-24 : I don't know why the callback
            # for debug: msg= doesn't put the hostname first like
            # other output. Undetermined if bug or feature.
            self.assertLogLineStartsWith(
                r"""\{'string': '\d.""", text)
            # ... handling loops is a different path, and that does
            self.assertLogLineStartsWith(
                r"""compute1 \| ok: \{'string': '\d.""", text)
            self.assertLogLine(
                r'fake \| skipping: Conditional result was False', text)

    @zuul_config('executor', 'output_max_bytes', '2')
    def test_command_output_max_bytes(self):
        job = self._run_job('command')
        with self.jobLog(job):
            build = self.history[-1]
            self.assertEqual(build.result, 'FAILURE')

            console_output = self.console_output.getvalue()
            self.assertNotIn('[WARNING]: Failure using method', console_output)
            text = self._get_job_output(build)
            self.assertIn('[Zuul] Log output exceeded max of 2', text)

    def test_module_exception(self):
        job = self._run_job('module_failure_exception')
        with self.jobLog(job):
            build = self.history[-1]
            self.assertEqual(build.result, 'FAILURE')

            text = self._get_job_output(build)
            self.assertLogLine(r'TASK \[Module failure\]', text)
            self.assertLogLine(
                r'controller \| MODULE FAILURE:', text)
            self.assertLogLine(
                r'controller \| Exception: This module is broken', text)

    def test_module_no_result(self):
        job = self._run_job('module_failure_no_result')
        with self.jobLog(job):
            build = self.history[-1]
            self.assertEqual(build.result, 'FAILURE')

            text = self._get_job_output(build)
            self.assertLogLine(r'TASK \[Module failure\]', text)

            regex = r'controller \|   "msg": "New-style module did not ' \
                r'handle its own exit"'
            self.assertLogLine(regex, text)

    # These twe tests are helpful to have for local debugging, but we
    # don't run them in the gate (yet) for two reasons:
    # 1) The VM test has no useful assertions since it was created as
    #    a comparison for the pod test
    # 2) The pod test can not be run in the gate.
    @skip("No useful assertions")
    def test_vm(self):
        # This test is not particularly useful, it is mostly a
        # benchmark to compare with the pod test below.
        nodes = [{'name': 'controller', 'label': 'whatever'}]
        job = self._run_job('vm', nodes=nodes)
        with self.jobLog(job):
            build = self.history[-1]
            path = os.path.join(self.jobdir_root, build.uuid,
                                'work', 'logs', 'job-output.txt')
            with open(path) as f:
                self.log.debug(f.read())

    @skip("Pod unavailable in gate")
    def test_pod(self):
        # This test could be used to verify the kubectl restart
        # functionality if we had a gate job with access to a pod.
        # TODO: add microk8s or kind to the stream test and enable
        # this
        nodes = [{'name': 'controller', 'label': 'remote-pod'}]
        job = self._run_job('pod', nodes=nodes)
        with self.jobLog(job):
            build = self.history[-1]
            path = os.path.join(self.jobdir_root, build.uuid,
                                'work', 'logs', 'job-output.txt')
            with open(path) as f:
                self.log.debug(f.read())

    @skip("Windows unavailable in gate")
    def test_win_command(self):
        self.fake_nodepool.shell_type = 'cmd'
        job = self._run_job('win-command')
        with self.jobLog(job):
            build = self.history[-1]
            self.assertEqual(build.result, 'SUCCESS')

            console_output = self.console_output.getvalue()
            # This should be generic enough to match any callback
            # plugin failures, which look something like
            #
            #  [WARNING]: Failure using method (v2_runner_on_ok) in \
            #                                                  callback plugin
            #  (<ansible.plugins.callback.zuul_stream.CallbackModule object at'
            #  0x7f89f72a20b0>): 'dict' object has no attribute 'startswith'"
            #  Callback Exception:
            #  ...
            #
            self.assertNotIn('[WARNING]: Failure using method', console_output)

            text = self._get_job_output(build)
            data = self._get_job_json(build)

            # The win_ modules do not have a strip_trailing_whitespace
            # option like the unix ones (and the unix ones default to
            # true), so we get a CRLF at the end of our stdout stream.
            token_stdout = "Standard output test {}\r\n".format(
                self.history[0].jobdir.src_root)
            # The win_shell module trims the stderr string as part of
            # its CLIXML handling, so there is no CRLF here.
            token_stderr = "Standard error test {}".format(
                self.history[0].jobdir.src_root)
            result = data[0]['plays'][1]['tasks'][2]['hosts']['compute1']
            self.assertEqual(token_stdout, result['stdout'])
            self.assertEqual(token_stderr, result['stderr'])

            # Find the "creates" tasks
            create1_task = data[0]['plays'][4]['tasks'][3]
            create1_host = create1_task['hosts']['compute1']
            self.assertIsNotNone(create1_host['delta'])
            self.assertNotIn("skipped, since", create1_host.get('msg', ''))
            self.assertEqual("Creates file that does not exist",
                             create1_task['task']['name'])
            create2_task = data[0]['plays'][4]['tasks'][4]
            create2_host = create2_task['hosts']['compute1']
            self.assertIsNone(create2_host.get('delta'))
            self.assertIn("skipped, since", create2_host['msg'])
            self.assertEqual("Creates file that already exists",
                             create2_task['task']['name'])
            # There is no "delta" returned in this case, so we don't
            # get a result linee
            # self.assertLogLine(r'compute1 \| ok: Runtime: None', text)

            # Test no_log for win_command
            nolog_task = data[0]['plays'][1]['tasks'][14]
            self.assertEqual("Test no_log command",
                             nolog_task['task']['name'])
            nolog_json_as_str = json.dumps(nolog_task)
            self.assertNotIn("This task should not log", nolog_json_as_str)
            self.assertNotIn('This task should not log', text)

            # Test no_log for win_shell
            nolog_task = data[0]['plays'][1]['tasks'][15]
            self.assertEqual("Test no_log shell",
                             nolog_task['task']['name'])
            nolog_json_as_str = json.dumps(nolog_task)
            self.assertNotIn("This task should not log", nolog_json_as_str)
            self.assertNotIn('This task should not log', text)

            # Test it does not appear in the registered output from command
            nolog_task = data[0]['plays'][1]['tasks'][16]
            self.assertEqual("Inspect command_no_log_result",
                             nolog_task['task']['name'])
            nolog_json_as_str = json.dumps(nolog_task)
            self.assertNotIn("This task should not log", nolog_json_as_str)

            # Test it does not appear in the registered output from shell
            nolog_task = data[0]['plays'][1]['tasks'][17]
            self.assertEqual("Inspect shell_no_log_result",
                             nolog_task['task']['name'])
            nolog_json_as_str = json.dumps(nolog_task)
            self.assertNotIn("This task should not log", nolog_json_as_str)

            self.assertLogLine(
                r'RUN START: \[untrusted : review.example.com/org/project/'
                r'playbooks/win-command.yaml@master\]', text)
            self.assertLogLine(r'PLAY \[all\]', text)
            self.assertLogLine(
                r'Ansible version={}'.format(self.ansible_core_version), text)
            self.assertLogLine(r'TASK \[Show contents of first file\]', text)
            self.assertLogLine(r'controller \| command test one', text)
            self.assertLogLine(
                r'controller \| ok: Runtime: \d:\d\d:\d\d\.\d\d\d\d\d\d', text)
            self.assertLogLine(r'TASK \[Show contents of second file\]', text)
            self.assertLogLine(r'compute1 \| command test two', text)
            self.assertLogLine(r'controller \| command test two', text)
            self.assertLogLine(r'compute1 \| This is a rescue task', text)
            self.assertLogLine(r'controller \| This is a rescue task', text)
            self.assertLogLine(r'compute1 \| This is an always task', text)
            self.assertLogLine(r'controller \| This is an always task', text)
            self.assertLogLine(r'compute1 \| This is a handler', text)
            self.assertLogLine(r'controller \| This is a handler', text)
            self.assertLogLine(r'controller \| First free task', text)
            self.assertLogLine(r'controller \| Second free task', text)
            self.assertLogLine(r'controller \| This is a shell task after an '
                               'included role', text)
            self.assertLogLine(r'compute1 \| This is a shell task after an '
                               'included role', text)
            self.assertLogLine(r'controller \| This is a command task after '
                               'an included role', text)
            self.assertLogLine(r'compute1 \| This is a command task after an '
                               'included role', text)
            self.assertLogLine(r'controller \| This is a shell task with '
                               'delegate compute1', text)
            self.assertLogLine(r'controller \| This is a shell task with '
                               'delegate controller', text)
            self.assertLogLine(r'compute1 \| item_in_loop1', text)
            self.assertLogLine(r'compute1 \| ok: Item: item_in_loop1 '
                               r'Runtime: \d:\d\d:\d\d\.\d\d\d\d\d\d', text)
            self.assertLogLine(r'compute1 \| item_in_loop2', text)
            self.assertLogLine(r'compute1 \| ok: Item: item_in_loop2 '
                               r'Runtime: \d:\d\d:\d\d\.\d\d\d\d\d\d', text)
            self.assertLogLine(r'compute1 \| failed_in_loop1', text)
            self.assertLogLine(r'compute1 \| ok: Item: failed_in_loop1 '
                               r'Result: 1', text)
            self.assertLogLine(r'compute1 \| failed_in_loop2', text)
            self.assertLogLine(r'compute1 \| ok: Item: failed_in_loop2 '
                               r'Result: 1', text)
            self.assertLogLine(r'compute1 \| transitive-one', text)
            self.assertLogLine(r'compute1 \| transitive-two', text)
            self.assertLogLine(r'compute1 \| transitive-three', text)
            self.assertLogLine(r'compute1 \| transitive-four', text)
            self.assertLogLine(
                r'controller \| ok: Runtime: \d:\d\d:\d\d\.\d\d\d\d\d\d', text)
            self.assertLogLine('PLAY RECAP', text)
            self.assertLogLine(
                r'controller \| ok: \d+ changed: \d+ unreachable: 0 failed: 0 '
                'skipped: 2 rescued: 1 ignored: 0', text)
            self.assertLogLine(
                r'RUN END RESULT_NORMAL: \[untrusted : review.example.com/'
                r'org/project/playbooks/win-command.yaml@master]', text)
            time1, time2 = self._getLogTime(r'TASK \[Command Not Found\]',
                                            text)
            self.assertLess((time2 - time1) / timedelta(milliseconds=1),
                            9000)

            # This is from the debug: msg='{{ ansible_version }}'
            # testing raw variable output.  To make it version
            # agnostic, match just the start of
            #  compute1 | ok: {'string': '2.9.27'...

            # NOTE(ianw) 2022-08-24 : I don't know why the callback
            # for debug: msg= doesn't put the hostname first like
            # other output. Undetermined if bug or feature.
            self.assertLogLineStartsWith(
                r"""\{'string': '\d.""", text)
            # ... handling loops is a different path, and that does
            self.assertLogLineStartsWith(
                r"""compute1 \| ok: \{'string': '\d.""", text)
            self.assertLogLine(
                r'fake \| skipping: Conditional result was False', text)

    @skip("Windows unavailable in gate")
    def test_win_command_fqcn(self):
        self.fake_nodepool.shell_type = 'cmd'
        job = self._run_job('win-command-fqcn')
        with self.jobLog(job):
            build = self.history[-1]
            self.assertEqual(build.result, 'SUCCESS')

            console_output = self.console_output.getvalue()
            # This should be generic enough to match any callback
            # plugin failures, which look something like
            #
            #  [WARNING]: Failure using method (v2_runner_on_ok) in \
            #                                                  callback plugin
            #  (<ansible.plugins.callback.zuul_stream.CallbackModule object at'
            #  0x7f89f72a20b0>): 'dict' object has no attribute 'startswith'"
            #  Callback Exception:
            #  ...
            #
            self.assertNotIn('[WARNING]: Failure using method', console_output)

            text = self._get_job_output(build)
            data = self._get_job_json(build)

            # The win_ modules do not have a strip_trailing_whitespace
            # option like the unix ones (and the unix ones default to
            # true), so we get a CRLF at the end of our stdout stream.
            token_stdout = "Standard output test {}\r\n".format(
                self.history[0].jobdir.src_root)
            # The win_shell module trims the stderr string as part of
            # its CLIXML handling, so there is no CRLF here.
            token_stderr = "Standard error test {}".format(
                self.history[0].jobdir.src_root)
            result = data[0]['plays'][1]['tasks'][2]['hosts']['compute1']
            self.assertEqual(token_stdout, result['stdout'])
            self.assertEqual(token_stderr, result['stderr'])

            # Find the "creates" tasks
            create1_task = data[0]['plays'][4]['tasks'][3]
            create1_host = create1_task['hosts']['compute1']
            self.assertIsNotNone(create1_host['delta'])
            self.assertNotIn("skipped, since", create1_host.get('msg', ''))
            self.assertEqual("Creates file that does not exist",
                             create1_task['task']['name'])
            create2_task = data[0]['plays'][4]['tasks'][4]
            create2_host = create2_task['hosts']['compute1']
            self.assertIsNone(create2_host.get('delta'))
            self.assertIn("skipped, since", create2_host['msg'])
            self.assertEqual("Creates file that already exists",
                             create2_task['task']['name'])
            # There is no "delta" returned in this case, so we don't
            # get a result linee
            # self.assertLogLine(r'compute1 \| ok: Runtime: None', text)

            # Test no_log for win_command
            nolog_task = data[0]['plays'][1]['tasks'][14]
            self.assertEqual("Test no_log command",
                             nolog_task['task']['name'])
            nolog_json_as_str = json.dumps(nolog_task)
            self.assertNotIn("This task should not log", nolog_json_as_str)
            self.assertNotIn('This task should not log', text)

            # Test no_log for win_shell
            nolog_task = data[0]['plays'][1]['tasks'][15]
            self.assertEqual("Test no_log shell",
                             nolog_task['task']['name'])
            nolog_json_as_str = json.dumps(nolog_task)
            self.assertNotIn("This task should not log", nolog_json_as_str)
            self.assertNotIn('This task should not log', text)

            self.assertLogLine(
                r'RUN START: \[untrusted : review.example.com/org/project/'
                r'playbooks/win-command-fqcn.yaml@master\]', text)
            self.assertLogLine(r'PLAY \[all\]', text)
            self.assertLogLine(
                r'Ansible version={}'.format(self.ansible_core_version), text)
            self.assertLogLine(r'TASK \[Show contents of first file\]', text)
            self.assertLogLine(r'controller \| command test one', text)
            self.assertLogLine(
                r'controller \| ok: Runtime: \d:\d\d:\d\d\.\d\d\d\d\d\d', text)
            self.assertLogLine(r'TASK \[Show contents of second file\]', text)
            self.assertLogLine(r'compute1 \| command test two', text)
            self.assertLogLine(r'controller \| command test two', text)
            self.assertLogLine(r'compute1 \| This is a rescue task', text)
            self.assertLogLine(r'controller \| This is a rescue task', text)
            self.assertLogLine(r'compute1 \| This is an always task', text)
            self.assertLogLine(r'controller \| This is an always task', text)
            self.assertLogLine(r'compute1 \| This is a handler', text)
            self.assertLogLine(r'controller \| This is a handler', text)
            self.assertLogLine(r'controller \| First free task', text)
            self.assertLogLine(r'controller \| Second free task', text)
            self.assertLogLine(r'controller \| This is a shell task after an '
                               'included role', text)
            self.assertLogLine(r'compute1 \| This is a shell task after an '
                               'included role', text)
            self.assertLogLine(r'controller \| This is a command task after '
                               'an included role', text)
            self.assertLogLine(r'compute1 \| This is a command task after an '
                               'included role', text)
            self.assertLogLine(r'controller \| This is a shell task with '
                               'delegate compute1', text)
            self.assertLogLine(r'controller \| This is a shell task with '
                               'delegate controller', text)
            self.assertLogLine(r'compute1 \| item_in_loop1', text)
            self.assertLogLine(r'compute1 \| ok: Item: item_in_loop1 '
                               r'Runtime: \d:\d\d:\d\d\.\d\d\d\d\d\d', text)
            self.assertLogLine(r'compute1 \| item_in_loop2', text)
            self.assertLogLine(r'compute1 \| ok: Item: item_in_loop2 '
                               r'Runtime: \d:\d\d:\d\d\.\d\d\d\d\d\d', text)
            self.assertLogLine(r'compute1 \| failed_in_loop1', text)
            self.assertLogLine(r'compute1 \| ok: Item: failed_in_loop1 '
                               r'Result: 1', text)
            self.assertLogLine(r'compute1 \| failed_in_loop2', text)
            self.assertLogLine(r'compute1 \| ok: Item: failed_in_loop2 '
                               r'Result: 1', text)
            self.assertLogLine(r'compute1 \| transitive-one', text)
            self.assertLogLine(r'compute1 \| transitive-two', text)
            self.assertLogLine(r'compute1 \| transitive-three', text)
            self.assertLogLine(r'compute1 \| transitive-four', text)
            self.assertLogLine(
                r'controller \| ok: Runtime: \d:\d\d:\d\d\.\d\d\d\d\d\d', text)
            self.assertLogLine('PLAY RECAP', text)
            self.assertLogLine(
                r'controller \| ok: \d+ changed: \d+ unreachable: 0 failed: 0 '
                'skipped: 2 rescued: 1 ignored: 0', text)
            self.assertLogLine(
                r'RUN END RESULT_NORMAL: \[untrusted : review.example.com/'
                r'org/project/playbooks/win-command-fqcn.yaml@master]', text)
            time1, time2 = self._getLogTime(r'TASK \[Command Not Found\]',
                                            text)
            self.assertLess((time2 - time1) / timedelta(milliseconds=1),
                            9000)

            # This is from the debug: msg='{{ ansible_version }}'
            # testing raw variable output.  To make it version
            # agnostic, match just the start of
            #  compute1 | ok: {'string': '2.9.27'...

            # NOTE(ianw) 2022-08-24 : I don't know why the callback
            # for debug: msg= doesn't put the hostname first like
            # other output. Undetermined if bug or feature.
            self.assertLogLineStartsWith(
                r"""\{'string': '\d.""", text)
            # ... handling loops is a different path, and that does
            self.assertLogLineStartsWith(
                r"""compute1 \| ok: \{'string': '\d.""", text)
            self.assertLogLine(
                r'fake \| skipping: Conditional result was False', text)


class TestZuulStream9(AnsibleZuulTestCase, FunctionalZuulStreamMixIn):
    ansible_version = '9'
    ansible_core_version = '2.16'

    def setUp(self):
        super().setUp()
        self._setUp()


class TestZuulStream11(AnsibleZuulTestCase, FunctionalZuulStreamMixIn):
    ansible_version = '11'
    ansible_core_version = '2.18'

    def setUp(self):
        super().setUp()
        self._setUp()



# Copyright 2014 Rackspace Australia
# Copyright 2021-2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import logging


class BaseReporter(object, metaclass=abc.ABCMeta):
    """Base class for reporters.

    Defines the exact public methods that must be supplied.
    """

    log = logging.getLogger("zuul.reporter.BaseReporter")

    def __init__(self, driver, connection, config=None, parse_context=None):
        self.driver = driver
        self.connection = connection
        self.config = config or {}
        self._action = None

    def setAction(self, action):
        self._action = action

    @abc.abstractmethod
    def report(self, item, phase1=True, phase2=True):
        """Send the compiled report message

        Two-phase reporting may be enabled if one or the other of the
        `phase1` or `phase2` arguments is False.

        Phase1 should report everything except the actual merge action.
        Phase2 should report only the merge action.

        :arg phase1 bool: Whether to enable phase1 reporting
        :arg phase2 bool: Whether to enable phase2 reporting

        """

    def getSubmitAllowNeeds(self, manager):
        """Get a list of code review labels that are allowed to be
        "needed" in the submit records for a change, with respect
        to this queue.  In other words, the list of review labels
        this reporter itself is likely to set before submitting.

        :arg PipelineManager manager: The pipeline manager for the queue.
        """
        return []

    def postConfig(self):
        """Run tasks after configuration is reloaded"""

    def _configErrorMatchesChange(self, err, change):
        context = err.key.context
        if not context:
            return False
        if context.project_canonical_name != \
                change.project.canonical_name:
            return False
        if not hasattr(change, 'branch'):
            return False
        if context.branch != change.branch:
            return False
        return True

    def addConfigurationErrorComments(self, item, change, comments):
        """Add file comments for configuration errors.

        Updates the comments dictionary with additional file comments
        for any relevant configuration errors for the specified change.

        :arg QueueItem item: The queue item
        :arg Ref change: One of the item's changes to check
        :arg dict comments: a file comments dictionary

        """

        for err in item.getConfigErrors():
            context = err.key.context
            mark = err.key.mark
            if not self._configErrorMatchesChange(err, change):
                continue
            if not (mark and err.short_error):
                return False
            if context.path not in change.files:
                return False
            existing_comments = comments.setdefault(context.path, [])
            existing_comments.append(dict(line=mark.end_line,
                                          message=err.short_error,
                                          range=dict(
                                              start_line=mark.line + 1,
                                              start_character=mark.column,
                                              end_line=mark.end_line,
                                              end_character=mark.end_column)))

    def _getFileComments(self, item, change):
        """Get the file comments from the zuul_return value"""
        ret = {}
        for build in item.current_build_set.getBuilds():
            fc = build.result_data.get("zuul", {}).get("file_comments")
            if not fc:
                continue
            # Only consider comments for this change
            if change.cache_key not in build.job.all_refs:
                continue
            for fn, comments in fc.items():
                existing_comments = ret.setdefault(fn, [])
                existing_comments.extend(comments)
        self.addConfigurationErrorComments(item, change, ret)
        return ret

    def getFileComments(self, item, change):
        comments = self._getFileComments(item, change)
        self.filterComments(item, change, comments)
        return comments

    def filterComments(self, item, change, comments):
        """Filter comments for files in change

        Remove any comments for files which do not appear in the
        specified change.  Leave warning messages if this happens.

        :arg QueueItem item: The queue item
        :arg Change change: The change
        :arg dict comments: a file comments dictionary (modified in place)
        """

        for fn in list(comments.keys()):
            if change.commentable_files is not None:
                if fn not in change.commentable_files:
                    del comments[fn]
                    item.warning("Comments left for invalid file %s" % (fn,))
            elif fn not in change.files:
                del comments[fn]
                item.warning("Comments left for invalid file %s" % (fn,))

    def _getFormatter(self, action):
        format_methods = {
            'enqueue': self._formatItemReportEnqueue,
            'start': self._formatItemReportStart,
            'success': self._formatItemReportSuccess,
            'failure': self._formatItemReportFailure,
            'merge-conflict': self._formatItemReportMergeConflict,
            'merge-failure': self._formatItemReportMergeFailure,
            'config-error': self._formatItemReportConfigError,
            'no-jobs': self._formatItemReportNoJobs,
            'disabled': self._formatItemReportDisabled,
            'dequeue': self._formatItemReportDequeue,
        }
        return format_methods[action]

    def _formatItemReport(self, item, change=None,
                          with_jobs=True, action=None, short=False):
        """Format a report from the given items. Usually to provide results to
        a reporter taking free-form text."""
        action = action or self._action
        ret = self._getFormatter(action)(item, change, with_jobs, short)

        config_warnings = item.getConfigErrors(errors=False, warnings=True)
        if config_warnings:
            ret += '\nWarning:\n  ' + config_warnings[0].error + '\n'

        if item.current_build_set.warning_messages:
            warning = '\n  '.join(item.current_build_set.warning_messages)
            ret += '\nWarning:\n  ' + warning + '\n'

        if item.current_build_set.debug_messages:
            debug = '\n  '.join(item.current_build_set.debug_messages)
            ret += '\nDebug information:\n  ' + debug + '\n'

        if item.manager.pipeline.footer_message:
            ret += '\n' + item.manager.pipeline.footer_message

        return ret

    def _formatItemReportEnqueue(self, item, change, with_jobs, short):
        return item.manager.pipeline.enqueue_message.format(
            pipeline=item.manager.pipeline.getSafeAttributes(),
            item_url=item.formatItemUrl())

    def _formatItemReportStart(self, item, change, with_jobs, short):
        return item.manager.pipeline.start_message.format(
            pipeline=item.manager.pipeline.getSafeAttributes(),
            item_url=item.formatItemUrl())

    def _formatItemReportSuccess(self, item, change, with_jobs, short):
        msg = item.manager.pipeline.success_message
        if with_jobs:
            item_url = item.formatItemUrl()
            if item_url is not None:
                msg += '\n' + item_url
            msg += '\n\n' + self._formatItemReportJobs(item, short)
        return msg

    def _formatItemReportFailure(self, item, change, with_jobs, short):
        if len(item.changes) > 1:
            _this_change = 'These changes'
            _depends = 'depend'
            _is = 'are'
        else:
            _this_change = 'This change'
            _depends = 'depends'
            _is = 'is'

        if item.dequeued_needing_change:
            msg = (f'{_this_change} {_depends} on a change '
                   'that failed to merge.\n')
            if isinstance(item.dequeued_needing_change, str):
                msg += '\n' + item.dequeued_needing_change + '\n'
        elif item.dequeued_missing_requirements:
            msg = (f'{_this_change} {_is} unable to merge '
                   'due to a missing merge requirement.\n')
        elif len(item.changes) > 1:
            # No plural phrasing here; we are specifically talking
            # about this change as part of a cycle.
            msg = ('This change is part of a dependency cycle '
                   'that failed.\n\n')

            # Attempt to provide relevant config error information
            # since we will not add it below.
            changes_with_errors = self._getChangesWithErrors(item)
            if change in changes_with_errors:
                msg += str(item.getConfigErrors(
                    errors=True, warnings=False)[0].error)
            change_annotations = {c: ' (config error)'
                                  for c in changes_with_errors}
            if with_jobs:
                msg = '{}\n{}'.format(msg, self._formatItemReportJobs(
                    item, short))
            msg = "{}\n{}".format(
                msg, self._formatItemReportOtherChanges(item,
                                                        change_annotations))
        elif item.didMergerFail():
            msg = item.manager.pipeline.merge_conflict_message
        elif item.current_build_set.has_blocking_errors:
            msg = str(item.getConfigErrors(
                errors=True, warnings=False)[0].error)
        else:
            msg = item.manager.pipeline.failure_message
            if with_jobs:
                item_url = item.formatItemUrl()
                if item_url is not None:
                    msg += '\n' + item_url
                msg += '\n\n' + self._formatItemReportJobs(item, short)
        return msg

    def _getChangesWithErrors(self, item):
        # Attempt to determine whether this change is the source of a
        # configuration error.  This is best effort.
        if not item.current_build_set.has_blocking_errors:
            return []

        ret = []
        for change in item.changes:
            for err in item.getConfigErrors():
                if self._configErrorMatchesChange(err, change):
                    ret.append(change)
        return ret

    def _formatItemReportMergeConflict(self, item, change, with_jobs, short):
        return item.manager.pipeline.merge_conflict_message

    def _formatItemReportMergeFailure(self, item, change, with_jobs, short):
        return 'This change was not merged by the code review system.\n'

    def _formatItemReportConfigError(self, item, change, with_jobs, short):
        if item.getConfigErrors():
            msg = str(item.getConfigErrors()[0].error)
        else:
            msg = "Unknown configuration error"
        return msg

    def _formatItemReportNoJobs(self, item, change, with_jobs, short):
        return item.manager.pipeline.no_jobs_message.format(
            pipeline=item.manager.pipeline.getSafeAttributes(),
            item_url=item.formatItemUrl())

    def _formatItemReportDisabled(self, item, change, with_jobs=True,
                                  short=False):
        if item.current_build_set.result == 'SUCCESS':
            return self._formatItemReportSuccess(item, change,
                                                 with_jobs, short)
        elif item.current_build_set.result == 'FAILURE':
            return self._formatItemReportFailure(item, change,
                                                 with_jobs, short)
        else:
            return self._formatItemReport(item, change, with_jobs, short)

    def _formatItemReportDequeue(self, item, change, with_jobs, short):
        msg = item.manager.pipeline.dequeue_message
        if with_jobs:
            msg += '\n\n' + self._formatItemReportJobs(item, short)
        return msg

    def _formatItemReportOtherChanges(self, item, change_annotations):
        change_lines = []
        for change in item.changes:
            annotation = change_annotations.get(change, '')
            change_lines.append(f'  - {change.url}{annotation}')
        return "Related changes:\n{}\n".format("\n".join(change_lines))

    def _getItemReportJobsFields(self, item):
        # Extract the report elements from an item
        config = self.connection.sched.config
        jobs_fields = []
        skipped = 0
        for job in item.getJobs():
            build = item.current_build_set.getBuild(job)
            (result, url) = item.formatJobResult(job)
            # If child_jobs is being used to skip jobs, then the user
            # probably has an expectation that some jobs will be
            # skipped and doesn't need to see all of them.  Otherwise,
            # it may be a surprise and it may be better to include the
            # job in the report.
            if (build.error_detail and
                'Skipped due to child_jobs' in build.error_detail):
                skipped += 1
                continue
            if not job.voting:
                voting = ' (non-voting)'
            else:
                voting = ''

            if config and config.has_option(
                'zuul', 'report_times'):
                report_times = config.getboolean(
                    'zuul', 'report_times')
            else:
                report_times = True

            if report_times and build.end_time and build.start_time:
                dt = int(build.end_time - build.start_time)
                m, s = divmod(dt, 60)
                h, m = divmod(m, 60)
                if h:
                    elapsed = ' in %dh %02dm %02ds' % (h, m, s)
                elif m:
                    elapsed = ' in %dm %02ds' % (m, s)
                else:
                    elapsed = ' in %ds' % (s)
            else:
                elapsed = ''
            if build.error_detail:
                error = ' ' + build.error_detail
            else:
                error = ''
            name = job.name + ' '
            # TODO(mordred) The gerrit consumption interface depends on
            # something existing in the url field and don't have a great
            # behavior defined for url being none/missing. Put name into
            # the url field to match old behavior until we can deal with
            # the gerrit-side piece as well
            url = url or job.name
            # Pass user defined success_message deciding the build result
            # during result formatting
            success_message = job.success_message
            jobs_fields.append(
                (name, url, result, error, elapsed, voting, success_message))
        return jobs_fields, skipped

    def _formatItemReportJobs(self, item, short):
        # Return the list of jobs portion of the report
        # If short is true, omit successful jobs for brevity
        ret = ''
        jobs_fields, skipped = self._getItemReportJobsFields(item)
        for job_fields in jobs_fields:
            if short and job_fields[2] == 'SUCCESS':
                continue
            ret += '- %s%s : %s%s%s%s\n' % job_fields[:6]
        if skipped:
            jobtext = 'job' if skipped == 1 else 'jobs'
            ret += 'Skipped %i %s\n' % (skipped, jobtext)
        if short:
            ret += 'Successful jobs omitted due to length restrictions\n'
        return ret



# Copyright 2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json
import argparse
import os
import subprocess
import logging
import sys

parser = argparse.ArgumentParser(
    description="Replay the Zuul workspace repo setup using information "
    "saved from a Zuul build's workspace-repos.json file.")
parser.add_argument('file', help='Path to workspace-repos.json')
parser.add_argument('-v', action='store_true', help='Verbose logging')
parser.add_argument('--workspace', help='Path to workspace')
options = parser.parse_args()

if options.v:
    logging.basicConfig(level=logging.DEBUG)
else:
    logging.basicConfig(level=logging.INFO)


class CloneException(Exception):
    pass


class Workspace:
    def __init__(self, path):
        self.path = path
        self.log = logging.getLogger('workspace')

    def run(self, cmd, cwd, timestamp=None, level=logging.INFO):
        self.log.log(level, 'Run: "%s"', ' '.join(cmd))
        env = self.env
        if timestamp:
            env = env.copy()
            env['GIT_COMMITTER_DATE'] = str(int(timestamp)) + '+0000'
            env['GIT_AUTHOR_DATE'] = str(int(timestamp)) + '+0000'
        subprocess.run(cmd, cwd=cwd, env=env, check=True)

    def load(self, fn):
        with open(fn) as f:
            data = json.load(f)
        self.repo_state = data['repo_state']
        self.merge_ops = data['merge_ops']
        self.merge_name = data['merge_name']
        self.merge_email = data['merge_email']
        self.env = {
            'GIT_AUTHOR_NAME': self.merge_name,
            'GIT_AUTHOR_EMAIL': self.merge_email,
            'GIT_COMITTER_NAME': self.merge_name,
            'GIT_COMITTER_EMAIL': self.merge_email,
        }

    def clone(self):
        for repo, state in self.repo_state.items():
            path = os.path.join(self.path, repo)
            fail = False
            if not os.path.exists(path):
                self.log.error("Please clone the repo %s", repo)
                fail = True
            if fail:
                raise CloneException()

    def reset(self):
        for repo, state in self.repo_state.items():
            path = os.path.join(self.path, repo)
            for ref, sha in state.items():
                self.run(['git', 'update-ref', ref, sha],
                         cwd=path, level=logging.DEBUG)

    def replay(self):
        for op in self.merge_ops:
            if 'cmd' in op:
                self.run(op['cmd'],
                         cwd=os.path.join(self.path, op['path']),
                         timestamp=op.get('timestamp'))
            if 'comment' in op:
                self.log.info(op['comment'])


w = Workspace(options.workspace or os.getcwd())
w.load(options.file)
try:
    w.clone()
except CloneException:
    sys.exit(1)
w.reset()
w.replay()



# Copyright 2021 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import json
import logging
import uuid

from kazoo.exceptions import NoNodeError, NodeExistsError

from zuul.zk import ZooKeeperBase


# This is outside of the /zuul root, so this holds data we expect to
# be persistent (it will not be deleted by 'zuul delete-state').
SYSTEM_ROOT = "/zuul-system"


class ZuulSystem(ZooKeeperBase):
    """Information about the complete Zuul system

    Currently includes only a system UUID for use with Nodepool.
    """

    log = logging.getLogger("zuul.System")

    def __init__(self, client):
        super().__init__(client)
        try:
            data, stat = self.kazoo_client.get(SYSTEM_ROOT)
        except NoNodeError:
            system_id = uuid.uuid4().hex
            data = json.dumps({'system_id': system_id},
                              sort_keys=True).encode('utf8')
            try:
                self.kazoo_client.create(SYSTEM_ROOT, data)
            except NodeExistsError:
                data, stat = self.kazoo_client.get(SYSTEM_ROOT)

        self.system_id = json.loads(data.decode('utf8'))['system_id']



# Copyright 2011 OpenStack, LLC.
# Copyright 2012 Hewlett-Packard Development Company, L.P.
# Copyright 2013 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import logging
import math
import threading
import time

from zuul.zk.event_queues import EventReceiverElection


class GerritChecksPoller(threading.Thread):
    # Poll gerrit without stream-events
    log = logging.getLogger("zuul.GerritConnection.poller")
    poll_interval = 30

    def __init__(self, connection):
        threading.Thread.__init__(self)
        self.connection = connection
        self.last_merged_poll = 0
        self.poller_election = EventReceiverElection(
            connection.sched.zk_client, connection.connection_name, "poller")
        self._stopped = False
        self._stop_event = threading.Event()

    def _makePendingCheckEvent(self, change, uuid, check):
        return {'type': 'pending-check',
                'uuid': uuid,
                'change': {
                    'project': change['patch_set']['repository'],
                    'number': change['patch_set']['change_number'],
                },
                'patchSet': {
                    'number': change['patch_set']['patch_set_id'],
                }}

    def _makeChangeMergedEvent(self, change):
        """Make a simulated change-merged event

        Mostly for the benefit of scheduler reconfiguration.
        """
        rev = change['revisions'][change['current_revision']]
        return {'type': 'change-merged',
                'change': {
                    'project': change['project'],
                    'number': change['_number'],
                },
                'patchSet': {
                    'number': rev['_number'],
                }}

    def _poll_checkers(self):
        for checker in self.connection.watched_checkers:
            changes = self.connection.get(
                'plugins/checks/checks.pending/?'
                'query=checker:%s+(state:NOT_STARTED)' % checker)
            for change in changes:
                for uuid, check in change['pending_checks'].items():
                    event = self._makePendingCheckEvent(
                        change, uuid, check)
                    self.connection.addEvent(event)

    def _poll_merged_changes(self):
        now = datetime.datetime.utcnow()
        age = self.last_merged_poll
        if age:
            # Allow an extra 4 seconds for request time
            age = int(math.ceil((now - age).total_seconds())) + 4
        changes = self.connection.simpleQueryHTTP(
            "status:merged -age:%ss" % (age,))
        self.last_merged_poll = now
        for change in changes:
            event = self._makeChangeMergedEvent(change)
            self.connection.addEvent(event)

    def _poll(self):
        next_start = self._last_start + self.poll_interval
        self._stop_event.wait(max(next_start - time.time(), 0))
        if self._stopped or not self.poller_election.is_still_valid():
            return
        self._last_start = time.time()
        self._poll_checkers()
        if self.connection.event_source == self.connection.EVENT_SOURCE_NONE:
            self._poll_merged_changes()

    def _run(self):
        self._last_start = time.time()
        while not self._stopped and self.poller_election.is_still_valid():
            # during tests, a sub-class _poll method is used to send
            # notifications
            self._poll()

    def run(self):
        while not self._stopped:
            try:
                self.poller_election.run(self._run)
            except Exception:
                self.log.exception("Exception on Gerrit poll with %s:",
                                   self.connection.connection_name)
                time.sleep(1)

    def stop(self):
        self.log.debug("Stopping watcher")
        self._stopped = True
        self._stop_event.set()
        self.poller_election.cancel()



# Copyright 2022-2025 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import math

from zuul.lib.voluputil import (
    AsList,
    Nullable,
    Optional,
    Required,
    assemble,
)

import voluptuous as vs

from zuul.driver.kubernetes.kubernetesendpoint import (
    KubernetesCreateStateMachine,
    KubernetesDeleteStateMachine,
)
from zuul.model import QuotaInformation
from zuul.provider import (
    BaseProvider,
    BaseProviderFlavor,
    BaseProviderImage,
    BaseProviderLabel,
    BaseProviderSchema,
)


class KubernetesProviderImage(BaseProviderImage):
    def __init__(self, image_config, provider_config):
        super().__init__(image_config, provider_config)
        # Implement provider defaults
        if self.connection_type is None:
            self.connection_type = 'kubectl'


class KubernetesProviderFlavor(BaseProviderFlavor):
    pass


class KubernetesProviderLabel(BaseProviderLabel):
    kubernetes_pull_secrets = vs.Schema({
        Required('name'): str,
        Optional('namespace', default='default'): str,
    })

    kubernetes_label_schema = vs.Schema({
        Required('kind'): vs.Any('pod', 'namespace'),
        Required('spec'): dict,
    })

    kubernetes_label_inheritable_schema = vs.Schema({
        Optional('kind'): Nullable(vs.Any('pod', 'namespace')),
    })

    kubernetes_label_common_schema = vs.Schema({
        Optional('annotations'): Nullable(dict),
        Optional('image-pull-secrets', default=[]): AsList(
            kubernetes_pull_secrets),
    })

    inheritable_schema = assemble(
        BaseProviderLabel.inheritable_schema,
        kubernetes_label_inheritable_schema,
        kubernetes_label_common_schema,
    )

    schema = assemble(
        BaseProviderLabel.schema,
        kubernetes_label_schema,
        kubernetes_label_common_schema,
    )

    image_flavor_inheritable_schema = vs.Schema({})

    def __init__(self, label_config, provider_config):
        super().__init__(label_config, provider_config)
        self.host_key_checking = False


class KubernetesProviderSchema(BaseProviderSchema):
    def getLabelSchema(self):
        return KubernetesProviderLabel.schema

    def getImageSchema(self):
        return KubernetesProviderImage.schema

    def getFlavorSchema(self):
        return KubernetesProviderFlavor.schema

    def getInheritableLabelSchema(self):
        return KubernetesProviderLabel.inheritable_schema

    def getInheritableImageSchema(self):
        return KubernetesProviderImage.inheritable_schema

    def getInheritableZuulImageSchema(self):
        return KubernetesProviderImage.inheritable_zuul_schema

    def getInheritableCloudImageSchema(self):
        return KubernetesProviderImage.inheritable_cloud_schema

    def getInheritableFlavorSchema(self):
        return KubernetesProviderFlavor.inheritable_schema

    def getProviderSchema(self):
        schema = super().getProviderSchema()

        resource_limits = {
            'pods': int,
            'namespaces': int,
        }

        kubernetes_provider_schema = vs.Schema({
            Optional('resource-limits', default=dict()): resource_limits,
        })

        return assemble(
            schema,
            kubernetes_provider_schema,
        )


class KubernetesProvider(BaseProvider, subclass_id='kubernetes'):
    log = logging.getLogger("zuul.KubernetesProvider")
    schema = KubernetesProviderSchema().getProviderSchema()

    @property
    def endpoint(self):
        ep = getattr(self, '_endpoint', None)
        if ep:
            return ep
        self._set(_endpoint=self.getEndpoint())
        return self._endpoint

    def parseImage(self, image_config, provider_config, connection):
        return KubernetesProviderImage(image_config, provider_config)

    def parseFlavor(self, flavor_config, provider_config, connection):
        return KubernetesProviderFlavor(flavor_config, provider_config)

    def parseLabel(self, label_config, provider_config, connection):
        return KubernetesProviderLabel(label_config, provider_config)

    def getEndpoint(self):
        return self.driver.getEndpoint(self)

    def getCreateStateMachine(self, node, image_external_id, log):
        # TODO: decide on a method of producing a hostname
        # that is max 15 chars.
        hostname = f"np{node.uuid[:13]}"
        label = self.labels[node.label]
        flavor = self.flavors[label.flavor]
        image = self.images[label.image]
        return KubernetesCreateStateMachine(
            self.endpoint,
            node,
            hostname,
            label,
            flavor,
            image,
            log)

    def getDeleteStateMachine(self, node, log):
        return KubernetesDeleteStateMachine(self.endpoint, node, log)

    def listInstances(self):
        return self.endpoint.listInstances()

    def getEndpointLimits(self):
        return QuotaInformation(default=math.inf)

    def getQuotaForLabel(self, label):
        return self.endpoint.getQuotaForLabel(label)

    def refreshQuotaForLabel(self, label, update):
        pass

    def getNodeTags(self, system_id, label, node_uuid,
                    provider=None, request=None):
        tags = super().getNodeTags(system_id, label, node_uuid,
                                   provider, request)
        # So that we can disambiguate requests for namespaces and pods
        # (both of which create namespaces).
        tags['zuul_kubernetes_kind'] = label.kind
        return tags



# Copyright 2022 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

# Analyze the contents of the ZK tree (whether in ZK or a dump on the
# local filesystem) to identify large objects.

import argparse
import json
import os
import sys
import zlib

import kazoo.client


KB = 1024
MB = 1024**2
GB = 1024**3


def convert_human(size):
    if size >= GB:
        return f'{int(size / GB)}G'
    if size >= MB:
        return f'{int(size / MB)}M'
    if size >= KB:
        return f'{int(size / KB)}K'
    if size > 0:
        return f'{size}B'
    return '0'


def convert_null(size):
    return size


def unconvert_human(size):
    suffix = size[-1]
    val = size[:-1]
    if suffix in ['G', 'g']:
        return int(val) * GB
    if suffix in ['M', 'm']:
        return int(val) * MB
    if suffix in ['K', 'k']:
        return int(val) * KB
    return int(size)


class SummaryLine:
    def __init__(self, kind, path, size=0, zk_size=0):
        self.kind = kind
        self.path = path
        self.size = size
        self.zk_size = zk_size
        self.attrs = {}
        self.children = []

    @property
    def tree_size(self):
        return sum([x.tree_size for x in self.children] + [self.size])

    @property
    def zk_tree_size(self):
        return sum([x.zk_tree_size for x in self.children] + [self.zk_size])

    def add(self, child):
        self.children.append(child)

    def __str__(self):
        indent = 0
        return self.toStr(indent)

    def matchesLimit(self, limit, zk):
        if not limit:
            return True
        if zk:
            size = self.zk_size
        else:
            size = self.size
        if size >= limit:
            return True
        for child in self.children:
            if child.matchesLimit(limit, zk):
                return True
        return False

    def toStr(self, indent, depth=None, conv=convert_null, limit=0, zk=False):
        """Convert this item and its children to a str representation

        :param indent int: How many levels to indent
        :param depth int: How many levels deep to display
        :param conv func: A function to convert sizes to text
        :param limit int: Don't display items smaller than this
        :param zk bool: Whether to use the data size (False)
                        or ZK storage size (True)
        """
        if depth and indent >= depth:
            return ''
        if self.matchesLimit(limit, zk):
            attrs = ' '.join([f'{k}={conv(v)}' for k, v in self.attrs.items()])
            if attrs:
                attrs = ' ' + attrs
            if zk:
                size = conv(self.zk_size)
                tree_size = conv(self.zk_tree_size)
            else:
                size = conv(self.size)
                tree_size = conv(self.tree_size)
            ret = ('  ' * indent + f"{self.kind} {self.path} "
                   f"size={size} tree={tree_size}{attrs}\n")
            for child in self.children:
                ret += child.toStr(indent + 1, depth, conv, limit, zk)
        else:
            ret = ''
        return ret


class Data:
    def __init__(self, path, raw, zk_size=None, failed=False):
        self.path = path
        self.raw = raw
        self.failed = failed
        self.zk_size = zk_size or len(raw)
        if not failed:
            self.data = json.loads(raw)
        else:
            print(f"!!! {path} failed to load data")
            self.data = {}

    @property
    def size(self):
        return len(self.raw)


class Tree:
    def getNode(self, path):
        pass

    def listChildren(self, path):
        pass

    def listConnections(self):
        return self.listChildren('/zuul/cache/connection')

    def getBranchCache(self, connection):
        return self.getShardedNode(f'/zuul/cache/connection/{connection}'
                                   '/branches/data')

    def listCacheKeys(self, connection):
        return self.listChildren(f'/zuul/cache/connection/{connection}/cache')

    def getCacheKey(self, connection, key):
        return self.getNode(f'/zuul/cache/connection/{connection}/cache/{key}')

    def listCacheData(self, connection):
        return self.listChildren(f'/zuul/cache/connection/{connection}/data')

    def getCacheData(self, connection, key):
        return self.getShardedNode(f'/zuul/cache/connection/{connection}'
                                   f'/data/{key}')

    def listTenants(self):
        return self.listChildren('/zuul/tenant')

    def listPipelines(self, tenant):
        return self.listChildren(f'/zuul/tenant/{tenant}/pipeline')

    def getPipeline(self, tenant, pipeline):
        return self.getNode(f'/zuul/tenant/{tenant}/pipeline/{pipeline}')

    def getItems(self, tenant, pipeline):
        pdata = self.getPipeline(tenant, pipeline)
        for queue in pdata.data.get('queues', []):
            qdata = self.getNode(queue)
            for item in qdata.data.get('queue', []):
                idata = self.getNode(item)
                yield idata

    def listBuildsets(self, item):
        return self.listChildren(f'{item}/buildset')

    def getBuildset(self, item, buildset):
        return self.getNode(f'{item}/buildset/{buildset}')

    def listJobs(self, buildset):
        return self.listChildren(f'{buildset}/job')

    def getJob(self, buildset, job_name):
        return self.getNode(f'{buildset}/job/{job_name}')

    def listBuilds(self, buildset, job_name):
        return self.listChildren(f'{buildset}/job/{job_name}/build')

    def getBuild(self, buildset, job_name, build):
        return self.getNode(f'{buildset}/job/{job_name}/build/{build}')


class FilesystemTree(Tree):
    def __init__(self, root):
        self.root = root

    def getNode(self, path):
        path = path.lstrip('/')
        fullpath = os.path.join(self.root, path)
        if not os.path.exists(fullpath):
            return Data(path, '', failed=True)
        try:
            with open(os.path.join(fullpath, 'ZKDATA'), 'rb') as f:
                zk_data = f.read()
                data = zk_data
                try:
                    data = zlib.decompress(zk_data)
                except Exception:
                    pass
                return Data(path, data, zk_size=len(zk_data))
        except Exception:
            return Data(path, '', failed=True)

    def getShardedNode(self, path):
        path = path.lstrip('/')
        fullpath = os.path.join(self.root, path)
        if not os.path.exists(fullpath):
            return Data(path, '', failed=True)
        shards = sorted([x for x in os.listdir(fullpath)
                         if x != 'ZKDATA'])
        data = b''
        compressed_data_len = 0
        try:
            for shard in shards:
                with open(os.path.join(fullpath, shard, 'ZKDATA'), 'rb') as f:
                    compressed_data = f.read()
                    compressed_data_len += len(compressed_data)
                    data += zlib.decompress(compressed_data)
            return Data(path, data, zk_size=compressed_data_len)
        except Exception:
            return Data(path, data, failed=True)

    def listChildren(self, path):
        path = path.lstrip('/')
        fullpath = os.path.join(self.root, path)
        if not os.path.exists(fullpath):
            return []
        return [x for x in os.listdir(fullpath)
                if x != 'ZKDATA']


class ZKTree(Tree):
    def __init__(self, host, cert, key, ca):
        kwargs = {}
        if cert:
            kwargs['use_ssl'] = True
            kwargs['keyfile'] = key
            kwargs['certfile'] = cert
            kwargs['ca'] = ca
        self.client = kazoo.client.KazooClient(host, **kwargs)
        self.client.start()

    def getNode(self, path):
        path = path.lstrip('/')
        if not self.client.exists(path):
            return Data(path, '', failed=True)
        try:
            zk_data, _ = self.client.get(path)
            data = zk_data
            try:
                data = zlib.decompress(zk_data)
            except Exception:
                pass
            return Data(path, data, zk_size=len(zk_data))
        except Exception:
            return Data(path, '', failed=True)

    def getShardedNode(self, path):
        path = path.lstrip('/')
        if not self.client.exists(path):
            return Data(path, '', failed=True)
        shards = sorted(self.listChildren(path))
        data = b''
        compressed_data_len = 0
        try:
            for shard in shards:
                compressed_data, _ = self.client.get(os.path.join(path, shard))
                compressed_data_len += len(compressed_data)
                data += zlib.decompress(compressed_data)
            return Data(path, data, zk_size=compressed_data_len)
        except Exception:
            return Data(path, data, failed=True)

    def listChildren(self, path):
        path = path.lstrip('/')
        try:
            return self.client.get_children(path)
        except kazoo.client.NoNodeError:
            return []


class Analyzer:
    def __init__(self, args):
        if args.path:
            self.tree = FilesystemTree(args.path)
        else:
            self.tree = ZKTree(args.host, args.cert, args.key, args.ca)
        if args.depth is not None:
            self.depth = int(args.depth)
        else:
            self.depth = None
        if args.human:
            self.conv = convert_human
        else:
            self.conv = convert_null
        if args.limit:
            self.limit = unconvert_human(args.limit)
        else:
            self.limit = 0
        self.use_zk_size = args.zk_size

    def summarizeItem(self, item):
        # Start with an item
        item_summary = SummaryLine('Item', item.path, item.size, item.zk_size)
        buildsets = self.tree.listBuildsets(item.path)
        for bs_i, bs_id in enumerate(buildsets):
            # Add each buildset
            buildset = self.tree.getBuildset(item.path, bs_id)
            buildset_summary = SummaryLine(
                'Buildset', buildset.path,
                buildset.size, buildset.zk_size)
            item_summary.add(buildset_summary)

            # Some attributes are offloaded, gather them and include
            # the size.
            for x in ['merge_repo_state', 'extra_repo_state', 'files',
                      'config_errors']:
                if buildset.data.get(x):
                    node = self.tree.getShardedNode(buildset.data.get(x))
                    buildset_summary.attrs[x] = \
                        self.use_zk_size and node.zk_size or node.size
                    buildset_summary.size += node.size
                    buildset_summary.zk_size += node.zk_size

            jobs = self.tree.listJobs(buildset.path)
            for job_i, job_name in enumerate(jobs):
                # Add each job
                job = self.tree.getJob(buildset.path, job_name)
                job_summary = SummaryLine('Job', job.path,
                                          job.size, job.zk_size)
                buildset_summary.add(job_summary)

                # Handle offloaded job data
                for job_attr in ('artifact_data',
                                 'extra_variables',
                                 'group_variables',
                                 'host_variables',
                                 'secret_parent_data',
                                 'variables',
                                 'parent_data',
                                 'secrets'):
                    job_data = job.data.get(job_attr, None)
                    if job_data and job_data['storage'] == 'offload':
                        node = self.tree.getShardedNode(job_data['path'])
                        job_summary.attrs[job_attr] = \
                            self.use_zk_size and node.zk_size or node.size
                        job_summary.size += node.size
                        job_summary.zk_size += node.zk_size

                builds = self.tree.listBuilds(buildset.path, job_name)
                for build_i, build_id in enumerate(builds):
                    # Add each build
                    build = self.tree.getBuild(
                        buildset.path, job_name, build_id)
                    build_summary = SummaryLine(
                        'Build', build.path, build.size, build.zk_size)
                    job_summary.add(build_summary)

                    # Add the offloaded build attributes
                    result_len = 0
                    result_zk_len = 0
                    if build.data.get('_result_data'):
                        result_data = self.tree.getShardedNode(
                            build.data['_result_data'])
                        result_len += result_data.size
                        result_zk_len += result_data.zk_size
                    if build.data.get('_secret_result_data'):
                        secret_result_data = self.tree.getShardedNode(
                            build.data['_secret_result_data'])
                        result_len += secret_result_data.size
                        result_zk_len += secret_result_data.zk_size
                    build_summary.attrs['results'] = \
                        self.use_zk_size and result_zk_len or result_len
                    build_summary.size += result_len
                    build_summary.zk_size += result_zk_len
        sys.stdout.write(item_summary.toStr(0, self.depth, self.conv,
                                            self.limit, self.use_zk_size))

    def summarizePipelines(self):
        for tenant_name in self.tree.listTenants():
            for pipeline_name in self.tree.listPipelines(tenant_name):
                for item in self.tree.getItems(tenant_name, pipeline_name):
                    self.summarizeItem(item)

    def summarizeConnectionCache(self, connection_name):
        connection_summary = SummaryLine('Connection', connection_name, 0, 0)
        branch_cache = self.tree.getBranchCache(connection_name)
        branch_summary = SummaryLine(
            'Branch Cache', connection_name,
            branch_cache.size, branch_cache.zk_size)
        connection_summary.add(branch_summary)

        cache_key_summary = SummaryLine(
            'Change Cache Keys', connection_name, 0, 0)
        cache_key_summary.attrs['count'] = 0
        connection_summary.add(cache_key_summary)
        for key in self.tree.listCacheKeys(connection_name):
            cache_key = self.tree.getCacheKey(connection_name, key)
            cache_key_summary.size += cache_key.size
            cache_key_summary.zk_size += cache_key.zk_size
            cache_key_summary.attrs['count'] += 1

        cache_data_summary = SummaryLine(
            'Change Cache Data', connection_name, 0, 0)
        cache_data_summary.attrs['count'] = 0
        connection_summary.add(cache_data_summary)
        for key in self.tree.listCacheData(connection_name):
            cache_data = self.tree.getCacheData(connection_name, key)
            cache_data_summary.size += cache_data.size
            cache_data_summary.zk_size += cache_data.zk_size
            cache_data_summary.attrs['count'] += 1

        sys.stdout.write(connection_summary.toStr(
            0, self.depth, self.conv, self.limit, self.use_zk_size))

    def summarizeConnections(self):
        for connection_name in self.tree.listConnections():
            self.summarizeConnectionCache(connection_name)

    def summarize(self):
        self.summarizeConnections()
        self.summarizePipelines()


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--path',
                        help='Filesystem path for previously dumped data')
    parser.add_argument('--host',
                        help='ZK host string (exclusive with --path)')
    parser.add_argument('--cert', help='Path to TLS certificate')
    parser.add_argument('--key', help='Path to TLS key')
    parser.add_argument('--ca', help='Path to TLS CA cert')
    parser.add_argument('-d', '--depth', help='Limit depth when printing')
    parser.add_argument('-H', '--human', dest='human', action='store_true',
                        help='Use human-readable sizes')
    parser.add_argument('-l', '--limit', dest='limit',
                        help='Only print nodes greater than limit')
    parser.add_argument('-Z', '--zksize', dest='zk_size', action='store_true',
                        help='Use the possibly compressed ZK storage size '
                        'instead of plain data size')
    args = parser.parse_args()

    az = Analyzer(args)
    az.summarize()



# Copyright 2018 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import textwrap

from tests.base import AnsibleZuulTestCase


class FunctionalActionModulesMixIn:
    tenant_config_file = 'config/remote-action-modules/main.yaml'
    # This should be overriden in child classes.
    ansible_version = 'X'
    wait_timeout = 120

    def _setUp(self):
        self.fake_nodepool.remote_ansible = True

        ansible_remote = os.environ.get('ZUUL_REMOTE_IPV4')
        self.assertIsNotNone(ansible_remote)

    def _run_job(self, job_name, result, expect_error=None):
        # Keep the jobdir around so we can inspect contents if an
        # assert fails. It will be cleaned up anyway as it is contained
        # in a tmp dir which gets cleaned up after the test.
        self.executor_server.keep_jobdir = True

        # Output extra ansible info so we might see errors.
        self.executor_server.verbose = True
        conf = textwrap.dedent(
            """
            - job:
                name: {job_name}
                run: playbooks/{job_name}.yaml
                ansible-version: {version}
                roles:
                  - zuul: org/common-config
                nodeset:
                  nodes:
                    - name: controller
                      label: whatever

            - project:
                check:
                  jobs:
                    - {job_name}
            """.format(job_name=job_name, version=self.ansible_version))

        file_dict = {'zuul.yaml': conf}
        A = self.fake_gerrit.addFakeChange('org/project', 'master', 'A',
                                           files=file_dict)
        self.fake_gerrit.addEvent(A.getPatchsetCreatedEvent(1))
        self.waitUntilSettled()

        job = self.getJobFromHistory(job_name)
        with self.jobLog(job):
            build = self.history[-1]
            self.assertEqual(build.result, result)

            if expect_error:
                path = os.path.join(self.jobdir_root, build.uuid,
                                    'work', 'logs', 'job-output.txt')
                with open(path, 'r') as f:
                    self.assertIn(expect_error, f.read())

    def test_command_module(self):
        self._run_job('command-good', 'SUCCESS')

    def test_zuul_return_module(self):
        self._run_job('zuul_return-good', 'SUCCESS')

    def test_zuul_return_module_delegate_to(self):
        self._run_job('zuul_return-good-delegate', 'SUCCESS')

    def test_shell_module(self):
        self._run_job('shell-good', 'SUCCESS')


class TestActionModules9(AnsibleZuulTestCase, FunctionalActionModulesMixIn):
    ansible_version = '9'

    def setUp(self):
        super().setUp()
        self._setUp()


class TestActionModules11(AnsibleZuulTestCase, FunctionalActionModulesMixIn):
    ansible_version = '11'

    def setUp(self):
        super().setUp()
        self._setUp()



# Copyright 2018 BMW Car IT GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc


class SensorInterface(object, metaclass=abc.ABCMeta):
    """The sensor interface used by the load governor

    A sensor which provides load indicators for managing the load
    on an executor.

    """

    def __init__(self, statsd, base_key):
        """Initialize the sensor.

        :param statsd: the statsd object to use for reporting
        :param base_key: the base key to use for reporting
        """
        self.statsd = statsd
        self.base_key = base_key

    @abc.abstractmethod
    def isOk(self):
        """Report if current load is ok for accepting new jobs.

        :returns: Bool, str: True if we can accept new jobs, otherwise False
                  and a string for the log
        :rtype: Bool, str
        """
        pass



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
import copy
import logging.config
import json
import os

import yaml

_DEFAULT_JOB_LOGGING_CONFIG = {
    'version': 1,
    'formatters': {
        'plain': {'format': '%(message)s'},
        'result': {'format': 'RESULT %(message)s'},
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stdout',
            'level': 'INFO',
            'formatter': 'plain',
        },
        'result': {
            # result is returned to subprocess stdout and is used to pass
            # control information from the callback back to the executor
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stdout',
            'level': 'INFO',
            'formatter': 'result',
        },
        'jobfile': {
            # used by executor to emit log file
            'class': 'logging.FileHandler',
            'level': 'INFO',
            'formatter': 'plain',
        },
    },
    'loggers': {
        'zuul.executor.ansible.result': {
            'handlers': ['result'],
            'level': 'INFO',
        },
        'zuul.executor.ansible': {
            'handlers': ['jobfile'],
            'level': 'INFO',
        },
        'sqlalchemy.engine': {
            'handlers': ['console'],
            'level': 'WARN',
        },
        'alembic': {
            'handlers': ['console'],
            'level': 'WARN',
        },
    },
    'root': {'handlers': []},
}

_DEFAULT_SERVER_LOGGING_CONFIG = {
    'version': 1,
    'formatters': {
        'simple': {
            'class': 'zuul.lib.logutil.MultiLineFormatter',
            'format': '%(asctime)s %(levelname)s %(name)s: %(message)s'
        },
    },
    'handlers': {
        'console': {
            # Used for printing to stdout
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stdout',
            'level': 'INFO',
            'formatter': 'simple',
        },
    },
    'loggers': {
        'zuul': {
            'handlers': ['console'],
            'level': 'INFO',
        },
        'zuul.GerritConnection.io': {
            'handlers': ['console'],
            'level': 'WARN',
        },
        'sqlalchemy.engine': {
            'handlers': ['console'],
            'level': 'WARN',
        },
        'gerrit': {
            'handlers': ['console'],
            'level': 'INFO',
        },
        'alembic.runtime.migration': {
            'handlers': ['console'],
            'level': 'INFO',
        },
        'alembic': {
            'handlers': ['console'],
            'level': 'WARN',
        },
        'cachecontrol': {
            'handlers': ['console'],
            'level': 'INFO',
        },
        'cherrypy': {
            'handlers': ['console'],
            'level': 'WARN',
        },
    },
    'root': {'handlers': []},
}

_DEFAULT_SERVER_FILE_HANDLERS = {
    'normal': {
        # Used for writing normal log files
        'class': 'logging.handlers.WatchedFileHandler',
        # This will get set to something real by DictLoggingConfig.server
        'filename': '/var/log/zuul/{server}.log',
        'level': 'INFO',
        'formatter': 'simple',
    },
}


def _read_config_file(filename: str):
    if not os.path.exists(filename):
        raise ValueError("Unable to read logging config file at %s" % filename)

    if os.path.splitext(filename)[1] in ('.yml', '.yaml', '.json'):
        with open(filename, 'r') as f:
            return yaml.safe_load(f)
    return filename


def load_config(filename: str):
    config = _read_config_file(filename)
    if isinstance(config, dict):
        return DictLoggingConfig(config)
    return FileLoggingConfig(filename)


def load_job_config(filename: str):
    return JobLoggingConfig(_read_config_file(filename))


class LoggingConfig(object, metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def apply(self):
        """Apply the config information to the current logging config."""


class DictLoggingConfig(LoggingConfig, metaclass=abc.ABCMeta):

    def __init__(self, config):
        self._config = config

    def apply(self):
        logging.config.dictConfig(self._config)

    def writeJson(self, filename: str):
        with open(filename, 'w') as f:
            f.write(json.dumps(self._config, indent=2))


class JobLoggingConfig(DictLoggingConfig):

    def __init__(self, config=None, job_output_file=None):
        if not config:
            config = copy.deepcopy(_DEFAULT_JOB_LOGGING_CONFIG)
        super(JobLoggingConfig, self).__init__(config=config)
        if job_output_file:
            self.job_output_file = job_output_file

    def setDebug(self):
        # Set the level for zuul.executor.ansible to DEBUG
        self._config['loggers']['zuul.executor.ansible']['level'] = 'DEBUG'

    @property
    def job_output_file(self) -> str:
        return self._config['handlers']['jobfile']['filename']

    @job_output_file.setter
    def job_output_file(self, filename: str):
        self._config['handlers']['jobfile']['filename'] = filename


class ServerLoggingConfig(DictLoggingConfig):

    def __init__(self, config=None, server=None):
        if not config:
            config = copy.deepcopy(_DEFAULT_SERVER_LOGGING_CONFIG)
        super(ServerLoggingConfig, self).__init__(config=config)
        if server:
            self.server = server

    @property
    def server(self):
        return self._server

    @server.setter
    def server(self, server):
        self._server = server
        # Add the normal file handler. It's not included in the default
        # config above because we're templating out the filename. Also, we
        # only want to add the handler if we're actually going to use it.
        for name, handler in _DEFAULT_SERVER_FILE_HANDLERS.items():
            server_handler = copy.deepcopy(handler)
            server_handler['filename'] = server_handler['filename'].format(
                server=server)
            self._config['handlers'][name] = server_handler
        # Change everything configured to write to stdout to write to
        # log files instead.
        for logger in self._config['loggers'].values():
            if logger['handlers'] == ['console']:
                logger['handlers'] = ['normal']

    def setDebug(self):
        # Change level from INFO to DEBUG
        for section in ('handlers', 'loggers'):
            for handler in self._config[section].values():
                if handler.get('level') == 'INFO':
                    handler['level'] = 'DEBUG'


class FileLoggingConfig(LoggingConfig):

    def __init__(self, filename):
        self._filename = filename

    def apply(self):
        logging.config.fileConfig(self._filename)


if __name__ == '__main__':
    # Use this to emit a working logging output for testing zuul_stream
    # locally.
    config = JobLoggingConfig(
        job_output_file=os.path.abspath('job-output.txt'))
    config.writeJson('logging.json')



# Copyright (C) 2011-2013 OpenStack Foundation
# Copyright 2017 Red Hat
# Copyright 2022 Acme Gating, LLC
# Copyright 2022-2024 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools
import logging
import math
import operator
import time
import urllib.parse
from collections import UserDict
from concurrent.futures import ThreadPoolExecutor

import openstack
from keystoneauth1.exceptions.catalog import EndpointNotFound

from zuul import exceptions
from zuul.driver.openstack.openstackmodel import (
    OpenstackResource,
    OpenstackInstance,
)
from zuul.driver.util import (
    LazyExecutorTTLCache,
    RateLimiter,
    Timer,
)
from zuul.model import QuotaInformation
from zuul.provider import (
    BaseImageUploadJob,
    BaseProviderEndpoint,
    statemachine
)


CACHE_TTL = 10


def quota_from_flavor(flavor, label=None, volumes=None):
    args = dict(instances=1,
                cores=flavor.vcpus,
                ram=flavor.ram)
    if label and label.boot_from_volume:
        args['volumes'] = 1
        args['volume-gb'] = label.volume_size
    elif volumes:
        args['volumes'] = len(volumes)
        args['volume-gb'] = sum([v.size for v in volumes])
    return QuotaInformation(**args)


def quota_from_limits(compute, volume):
    def bound_value(value):
        if value == -1:
            return math.inf
        return value

    args = dict(
        instances=bound_value(compute.max_total_instances),
        cores=bound_value(compute.max_total_cores),
        ram=bound_value(compute.max_total_ram_size))
    if volume is not None:
        args['volumes'] = bound_value(volume['absolute']['maxTotalVolumes'])
        args['volume-gb'] = bound_value(
            volume['absolute']['maxTotalVolumeGigabytes'])
    return QuotaInformation(**args)


def is_message_about_quota(msg):
    msg = msg.lower()
    if 'quota' in msg:
        return True
    if 'ports exceeded' in msg:
        return True
    if all(s in msg for s in ('insufficient', 'resources')):
        # Exceeded maximum number of retries. Exceeded max scheduling
        # attempts 6 for instance
        # f9daa9f2-1181-4693-827f-ba84cd4b002c. Last exception:
        # Insufficient compute resources: Free vcpu 0.00 VCPU <
        # requested 8 VCPU.
        return True


class ZuulOpenstackServer(UserDict):
    # Most of OpenStackSDK is designed around a dictionary interface,
    # but due to the historic use of the Munch object, some code
    # (add_server_interfaces) accesses values by attribute instead of
    # key.  For just those values, we provide getters and setters.

    @property
    def access_ipv4(self):
        return self.data.get('access_ipv4')

    @access_ipv4.setter
    def access_ipv4(self, value):
        self.data['access_ipv4'] = value

    @property
    def public_v4(self):
        return self.data.get('public_v4')

    @public_v4.setter
    def public_v4(self, value):
        self.data['public_v4'] = value

    @property
    def private_v4(self):
        return self.data.get('private_v4')

    @private_v4.setter
    def private_v4(self, value):
        self.data['private_v4'] = value

    @property
    def access_ipv6(self):
        return self.data.get('access_ipv6')

    @access_ipv6.setter
    def access_ipv6(self, value):
        self.data['access_ipv6'] = value

    @property
    def public_v6(self):
        return self.data.get('public_v6')

    @public_v6.setter
    def public_v6(self, value):
        self.data['public_v6'] = value


class OpenstackDeleteStateMachine(statemachine.StateMachine):
    FLOATING_IP_DELETING = 'deleting floating ip'
    SERVER_DELETE = 'delete server'
    SERVER_DELETING = 'deleting server'
    COMPLETE = 'complete'

    def __init__(self, endpoint, node, log):
        self.log = log
        self.endpoint = endpoint
        self.node = node
        super().__init__(node.delete_state)
        self.delete_future = None

        # Restore local objects
        self.server = None
        self.floating_ips = []
        if self.node.openstack_server_id:
            self.server = self.endpoint._getServer(
                self.node.openstack_server_id)
            if (self.server and
                self.endpoint._hasFloatingIps() and
                self.server.get('addresses')):
                self.floating_ips = self.endpoint._getFloatingIps(
                    self.server)

    def advance(self):
        if self.state == self.START:
            if self.node.openstack_server_id:
                for fip in self.floating_ips:
                    self.endpoint._deleteFloatingIp(fip)
                    self.state = self.FLOATING_IP_DELETING
                if not self.floating_ips:
                    self.state = self.SERVER_DELETE
            else:
                self.state = self.COMPLETE

        if self.state == self.FLOATING_IP_DELETING:
            fips = []
            for fip in self.floating_ips:
                fip = self.endpoint._refreshFloatingIpDelete(fip)
                if not fip or fip['status'] == 'DOWN':
                    fip = None
                if fip:
                    fips.append(fip)
            self.floating_ips = fips
            if self.floating_ips:
                return
            else:
                self.state = self.SERVER_DELETE

        if self.state == self.SERVER_DELETE:
            if not self.delete_future:
                self.delete_future = self.endpoint._submitApi(
                    self.endpoint._deleteServer,
                    self.node.openstack_server_id)
            if self.endpoint._completeApi(self.delete_future):
                self.state = self.SERVER_DELETING

        if self.state == self.SERVER_DELETING:
            self.server = self.endpoint._refreshServerDelete(self.server)
            if self.server:
                return
            else:
                self.state = self.COMPLETE

        if self.state == self.COMPLETE:
            self.complete = True


class OpenstackCreateStateMachine(statemachine.StateMachine):
    SERVER_CREATING_SUBMIT = 'submit creating server'
    SERVER_CREATING = 'creating server'
    FLOATING_IP_CREATING = 'creating floating ip'
    FLOATING_IP_ATTACHING = 'attaching floating ip'
    COMPLETE = 'complete'

    def __init__(self, endpoint, node, hostname, label, flavor, image,
                 image_external_id, tags, log):
        self.log = log
        self.endpoint = endpoint
        self.node = node
        self.label = label
        self.flavor = flavor
        self.image = image
        self.hostname = hostname
        self.az = label.az
        self.create_future = None
        super().__init__(node.create_state)
        self.attempts = node.create_state.get("attempts", 0)
        self.image_external_id = node.create_state.get(
            "image_external_id", image_external_id)
        self.config_drive = image.config_drive

        if image_external_id:
            self.image_external = image_external_id
        else:
            # launch using unmanaged cloud image

            if image.image_id:
                # Using a dict with the ID bypasses an image search during
                # server creation.
                self.image_external = dict(id=image.image_id)
            else:
                self.image_external = image.external_name

        meta = {}
        meta.update(tags)
        self.metadata = meta
        self.os_flavor = self.endpoint._findFlavor(
            flavor_name=flavor.flavor_name,
        )
        self.node.quota = quota_from_flavor(self.os_flavor, label=self.label)

        if self.state == self.SERVER_CREATING_SUBMIT:
            for instance in self.endpoint.listInstances():
                if instance.metadata.get('zuul_node_uuid') == node.uuid:
                    self.node.openstack_server_id =\
                        instance.openstack_server_id
            if self.node.openstack_server_id:
                self.state = self.SERVER_CREATING

        self.server = None
        if self.node.openstack_server_id:
            self.server = self.endpoint._refreshServer(
                dict(
                    id=self.node.openstack_server_id,
                    status='_unknown',
                ))
        self.floating_ip = None
        if self.node.openstack_floating_ip_id:
            self.floating_ip = self.endpoint._refreshFloatingIp(
                dict(id=self.node.openstack_floating_ip_id))

    def _handleServerFault(self):
        # Return True if this is a quota fault
        if not self.node.openstack_server_id:
            return
        try:
            server = self.endpoint._getServerByIdNow(
                self.node.openstack_server_id)
            if not server:
                return
            fault = server.get('fault', {}).get('message')
            if fault:
                self.log.error('Detailed node error: %s', fault)
                if is_message_about_quota(fault):
                    return True
        except Exception:
            self.log.exception(
                'Failed to retrieve node error information:')

    def toDict(self):
        data = super().toDict()
        data.update(
            attempts=self.attempts,
            image_external_id=self.image_external_id,
        )
        return data

    def advance(self):
        if self.state == self.START:
            self.state = self.SERVER_CREATING_SUBMIT

        if self.state == self.SERVER_CREATING_SUBMIT:
            if not self.create_future:
                self.create_future = self.endpoint._submitApi(
                    self.endpoint._createServer,
                    self.hostname,
                    image=self.image_external,
                    flavor=self.os_flavor,
                    key_name=self.label.key_name,
                    az=self.az,
                    config_drive=self.config_drive,
                    networks=self.label.networks,
                    security_groups=self.label.security_groups,
                    boot_from_volume=self.label.boot_from_volume,
                    volume_size=self.label.volume_size,
                    instance_properties=self.metadata,
                    userdata=self.label.userdata,
                )
            try:
                try:
                    self.server = self.endpoint._completeApi(
                        self.create_future)
                    if self.server is None:
                        return
                    self.node.openstack_server_id = self.server['id']
                    self.state = self.SERVER_CREATING
                except openstack.cloud.exc.OpenStackCloudCreateException as e:
                    if e.resource_id:
                        self.node.openstack_server_id = e.resource_id
                        if self._handleServerFault():
                            self.log.error("Launch attempt failed: %s", str(e))
                            raise exceptions.QuotaException("Quota exceeded")
                        raise
            except Exception as e:
                if is_message_about_quota(str(e)):
                    self.log.error("Launch attempt failed: %s", str(e))
                    raise exceptions.QuotaException("Quota exceeded")
                raise

        if self.state == self.SERVER_CREATING:
            self.server = self.endpoint._refreshServer(self.server)

            if self.server['status'] == 'ACTIVE':
                if (self.label.auto_floating_ip and
                    self.endpoint._needsFloatingIp(self.server)):
                    self.floating_ip = self.endpoint._createFloatingIp(
                        self.server)
                    self.node.openstack_floating_ip_id = self.floating_ip['id']
                    self.state = self.FLOATING_IP_CREATING
                else:
                    self.state = self.COMPLETE
            elif self.server['status'] == 'ERROR':
                if ('fault' in self.server and self.server['fault'] is not None
                    and 'message' in self.server['fault']):
                    self.log.error(
                        "Error in creating the server."
                        " Compute service reports fault: {reason}".format(
                            reason=self.server['fault']['message']))
                    error_message = self.server['fault']['message']
                    if is_message_about_quota(error_message):
                        raise exceptions.QuotaException("Quota exceeded")
                raise exceptions.LaunchStatusException("Server in error state")
            else:
                return

        if self.state == self.FLOATING_IP_CREATING:
            self.floating_ip = self.endpoint._refreshFloatingIp(
                self.floating_ip)
            if self.floating_ip.get('port_id', None):
                if self.floating_ip['status'] == 'ACTIVE':
                    self.state = self.FLOATING_IP_ATTACHING
                else:
                    return
            else:
                self.endpoint._attachIpToServer(self.server, self.floating_ip)
                self.state = self.FLOATING_IP_ATTACHING

        if self.state == self.FLOATING_IP_ATTACHING:
            self.server = self.endpoint._refreshServer(self.server)
            ext_ip = openstack.cloud.meta.get_server_ip(
                self.server, ext_tag='floating', public=True)
            if ext_ip == self.floating_ip['floating_ip_address']:
                self.state = self.COMPLETE
            else:
                return

        if self.state == self.COMPLETE:
            self.complete = True
            return self.endpoint._getInstance(self.server, self.node.quota)


class OpenstackImageUploadJob(BaseImageUploadJob):
    def __init__(self, endpoint, provider_image, image_name,
                 image_format, metadata, md5, sha256):
        super().__init__()
        self.endpoint = endpoint
        self.provider_image = provider_image
        self.image_name = image_name
        self.image_format = image_format
        self.metadata = metadata
        self.md5 = md5
        self.sha256 = sha256

    def run(self, filename):
        self.endpoint.log.debug(f"Uploading image {self.image_name}")

        image_id = self.endpoint._uploadImage(
            filename,
            self.provider_image, self.image_name,
            self.image_format, self.metadata,
            self.md5, self.sha256)
        return image_id


class OpenstackProviderEndpoint(BaseProviderEndpoint):
    """An OPENSTACK Endpoint corresponds to a single OPENSTACK region,
    and can include multiple availability zones.
    """

    IMAGE_UPLOAD_SLEEP = 30

    def __init__(self, zk_client, driver, connection, region, system_id):
        name = f'{connection.connection_name}-{region}'
        super().__init__(zk_client, driver, connection, name, system_id)
        self.log = logging.getLogger(f"zuul.openstack.{self.name}")
        self.region = region

        # Wrap these instance methods with a per-instance LRU cache so
        # that we don't leak memory over time when the endpoint is
        # occasionally replaced.
        self._findImage = functools.lru_cache(maxsize=None)(
            self._findImage)
        self._listFlavors = functools.lru_cache(maxsize=None)(
            self._listFlavors)
        self._findNetwork = functools.lru_cache(maxsize=None)(
            self._findNetwork)
        self._listAZs = functools.lru_cache(maxsize=None)(
            self._listAZs)

        self.rate_limiter = RateLimiter(self.name, connection.rate)

        self._last_image_check_failure = time.time()
        self._last_port_cleanup = None

    def startEndpoint(self):
        self.log.debug("Starting OpenStack endpoint")
        self._client = self._getClient()
        self._running = True
        # The default http connection pool size is 10; match it for
        # efficiency.
        workers = 10
        self.log.info("Create executor with max workers=%s", workers)
        self.api_executor = ThreadPoolExecutor(
            thread_name_prefix=f'openstack-api-{self.name}',
            max_workers=workers)

        # Use a lazy TTL cache for these.  This uses the TPE to
        # asynchronously update the cached values, meanwhile returning
        # the previous cached data if available.  This means every
        # call after the first one is instantaneous.
        self._listServers = LazyExecutorTTLCache(
            CACHE_TTL, self.api_executor)(
                self._listServers)
        self._listVolumes = LazyExecutorTTLCache(
            CACHE_TTL, self.api_executor)(
                self._listVolumes)
        self._listFloatingIps = LazyExecutorTTLCache(
            CACHE_TTL, self.api_executor)(
                self._listFloatingIps)

    def stopEndpoint(self):
        self.log.debug("Stopping OpenStack endpoint")
        self.api_executor.shutdown()
        self._running = False

    def listResources(self, providers):
        for server in self._listServers():
            if server['status'].lower() == 'deleted':
                continue
            yield OpenstackResource(server.get('metadata', {}),
                                    OpenstackResource.TYPE_INSTANCE,
                                    server['id'])
        # Floating IP and port leakage can't be handled by the
        # automatic resource cleanup in cleanupLeakedResources because
        # openstack doesn't store metadata on those objects, so we
        # call internal cleanup methods here.
        intervals = [p.port_cleanup_interval for p in providers
                     if p.port_cleanup_interval]
        interval = min(intervals or [0])
        if interval:
            self._cleanupLeakedPorts(interval)
        if any([p.floating_ip_cleanup for p in providers]):
            self._cleanupFloatingIps()

    def deleteResource(self, resource):
        self.log.info(f"Deleting leaked {resource.type}: {resource.id}")
        if resource.type == OpenstackResource.TYPE_INSTANCE:
            self._deleteServer(resource.id)

    def listInstances(self):
        volumes = {}
        for volume in self._listVolumes():
            volumes[volume.id] = volume
        for server in self._listServers():
            if server['status'].lower() == 'deleted':
                continue
            flavor = self._getFlavorFromServer(server)
            server_volumes = []
            for vattach in server.get(
                    'os-extended-volumes:volumes_attached', []):
                volume = volumes.get(vattach['id'])
                if volume:
                    server_volumes.append(volume)
            quota = quota_from_flavor(flavor, volumes=server_volumes)
            yield OpenstackInstance(self.connection.cloud_name,
                                    self.getRegionName(), server, quota)

    def getQuotaLimits(self):
        with Timer(self.log, 'API call get_compute_limits'):
            compute = self._client.get_compute_limits()
        try:
            with Timer(self.log, 'API call get_volume_limits'):
                volume = self._client.get_volume_limits()
        except EndpointNotFound:
            volume = None
        return quota_from_limits(compute, volume)

    def refreshQuotaLimits(self, update):
        if self.quota_cache.hasLimits() and not update:
            return False
        limits = self.getQuotaLimits()
        self.quota_cache.setLimits(limits)
        return True

    def getQuotaForLabel(self, label, flavor):
        # This should not rely on the client connection, only the
        # quota cache.
        key = f'flavor-{flavor.flavor_name}'
        return self.quota_cache.getResource(key)

    def refreshQuotaForLabel(self, label, flavor, update):
        key = f'flavor-{flavor.flavor_name}'
        if update or not self.quota_cache.hasResource(key):
            os_flavor = self._findFlavorByName(flavor.flavor_name)
            quota = quota_from_flavor(os_flavor, label=label)
            self.quota_cache.setResource(key, quota)

    def getAZs(self):
        # TODO: This is currently unused; it's unclear if we will need
        # to incorporate this into the node reuse process in the
        # launcher.
        azs = self._listAZs()
        if not azs:
            # If there are no zones, return a list containing None so that
            # random.choice can pick None and pass that to Nova. If this
            # feels dirty, please direct your ire to policy.json and the
            # ability to turn off random portions of the OpenStack API.
            return [None]
        return azs

    def labelReady(self, label):
        if not label.cloud_image:
            return False

        # If an image ID was supplied, we'll assume it is ready since
        # we don't currently have a way of validating that (except during
        # server creation).
        if label.cloud_image.image_id:
            return True

        image = self._findImage(label.cloud_image.external_name)
        if not image:
            self.log.warning(
                "Provider %s is configured to use %s as the"
                " cloud-image for label %s and that"
                " cloud-image could not be found in the"
                " cloud." % (self.provider.name,
                             label.cloud_image.external_name,
                             label.name))
            # If the user insists there should be an image but it
            # isn't in our cache, invalidate the cache periodically so
            # that we can see new cloud image uploads.
            if (time.time() - self._last_image_check_failure >
                self.IMAGE_CHECK_TIMEOUT):
                self._findImage.cache_clear()
                self._last_image_check_failure = time.time()
            return False
        return True

    def getImageUploadJob(self, provider_image, image_name,
                          image_format, metadata, md5, sha256):
        return OpenstackImageUploadJob(
            self,
            provider_image, image_name,
            image_format, metadata, md5, sha256)

    def _uploadImage(self, filename, provider_image, image_name,
                     image_format, metadata, md5, sha256):
        # configure glance and upload image.  Note the meta flags
        # are provided as custom glance properties
        # NOTE: we have wait=True set here. This is not how we normally
        # do things in nodepool, preferring to poll ourselves thankyouverymuch.
        # However - two things to note:
        #  - PUT has no aysnc mechanism, so we have to handle it anyway
        #  - v2 w/task waiting is very strange and complex - but we have to
        #              block for our v1 clouds anyway, so we might as well
        #              have the interface be the same and treat faking-out
        #              a openstacksdk-level fake-async interface later
        timeout = provider_image.import_timeout
        if not metadata:
            metadata = {}
        if image_format:
            metadata['disk_format'] = image_format
        with Timer(self.log, 'API call create_image'):
            image = self._client.create_image(
                name=image_name,
                filename=filename,
                is_public=False,
                wait=True,
                md5=md5,
                sha256=sha256,
                timeout=timeout,
                **metadata)
        return image.id

    def deleteImage(self, external_id):
        self.log.debug(f"Deleting image {external_id}")
        with Timer(self.log, 'API call delete_image'):
            return self._client.delete_image(external_id)

    # Local implementation

    def _getInstance(self, server, quota):
        return OpenstackInstance(
            self.connection.cloud_name,
            self.getRegionName(),
            server, quota)

    def _getRegionConfig(self):
        config = openstack.config.OpenStackConfig(
            config_files=self.connection.config_files,
            load_envvars=False,
            app_name='zuul',
        )
        return config.get_one(cloud=self.connection.cloud_name,
                              region_name=self.region)

    def _getClient(self):
        region = self._getRegionConfig()
        # Note: the connection object holds resources internally that
        # prevent GC.  Currently our endpoints are long-lived, and we
        # don't perform any cleanup for them.  If we do in the future,
        # we will need to call Connection.close() to release
        # resources.
        return openstack.connection.Connection(
            config=region,
            use_direct_get=False,
            rate_limit=self.connection.rate,
        )

    def getImageFormat(self):
        region = self._getRegionConfig()
        return region.config['image_format']

    def getRegionName(self):
        # With OpenStackSDK, users can omit the region and the SDK may
        # supply a default region.  This helper method will return the
        # actual region in use regardless of what the user supplied.
        return self._client.config.config['region_name']

    def _submitApi(self, api, *args, **kw):
        return self.api_executor.submit(
            api, *args, **kw)

    def _completeApi(self, future):
        if not future.done():
            return None
        return future.result()

    def _createServer(self, name, image, flavor,
                      az=None, key_name=None, config_drive=True,
                      networks=None, security_groups=None,
                      boot_from_volume=False, volume_size=50,
                      instance_properties=None, userdata=None):
        if not networks:
            networks = []
        if not isinstance(image, dict):
            # if it's a dict, we already have the cloud id. If it's not,
            # we don't know if it's name or ID so need to look it up
            image = self._findImage(image)
        create_args = dict(name=name,
                           image=image,
                           flavor=flavor,
                           config_drive=config_drive)
        if boot_from_volume:
            create_args['boot_from_volume'] = boot_from_volume
            create_args['volume_size'] = volume_size
            # NOTE(pabelanger): Always cleanup volumes when we delete a server.
            create_args['terminate_volume'] = True
        if key_name:
            create_args['key_name'] = key_name
        if az:
            create_args['availability_zone'] = az
        if security_groups:
            create_args['security_groups'] = security_groups
        if userdata:
            create_args['userdata'] = userdata
        nics = []
        for network in networks:
            net_id = self._findNetwork(network)['id']
            nics.append({'net-id': net_id})
        if nics:
            create_args['nics'] = nics
        if instance_properties:
            create_args['meta'] = instance_properties

        try:
            with Timer(self.log, 'API call create_server'):
                return self._client.create_server(wait=False, **create_args)
        except openstack.exceptions.BadRequestException:
            # We've gotten a 400 error from nova - which means the request
            # was malformed. The most likely cause of that, unless something
            # became functionally and systemically broken, is stale az, image
            # or flavor cache. Log a message, invalidate the caches so that
            # next time we get new caches.
            self.log.info(
                "Clearing az, flavor and image caches due to 400 error "
                "from nova")
            self._findImage.cache_clear()
            self._listFlavors.cache_clear()
            self._findNetwork.cache_clear()
            self._listAZs.cache_clear()
            raise

    # This method is wrapped with an LRU cache in the constructor.
    def _listAZs(self):
        with Timer(self.log, 'API call list_availability_zone_names'):
            return self._client.list_availability_zone_names()

    # This method is wrapped with an LRU cache in the constructor.
    def _findImage(self, name):
        with Timer(self.log, 'API call get_image'):
            return self._client.get_image(name, filters={'status': 'active'})

    # This method is wrapped with an LRU cache in the constructor.
    def _listFlavors(self):
        with Timer(self.log, 'API call list_flavors'):
            flavors = self._client.list_flavors(get_extra=False)
        flavors.sort(key=operator.itemgetter('ram', 'name'))
        return flavors

    # This method is only used by the nodepool alien-image-list
    # command and only works with the openstack driver.
    def _listImages(self):
        with Timer(self.log, 'API call list_images'):
            return self._client.list_images()

    def _findFlavorByName(self, flavor_name):
        for f in self._listFlavors():
            if flavor_name in (f['name'], f['id']):
                return f
        raise Exception("Unable to find flavor: %s" % flavor_name)

    def _findFlavorByRam(self, min_ram, flavor_name):
        for f in self._listFlavors():
            if (f['ram'] >= min_ram
                    and (not flavor_name or flavor_name in f['name'])):
                return f
        raise Exception("Unable to find flavor with min ram: %s" % min_ram)

    def _findFlavorById(self, flavor_id):
        for f in self._listFlavors():
            if f['id'] == flavor_id:
                return f
        raise Exception("Unable to find flavor with id: %s" % flavor_id)

    def _findFlavor(self, flavor_name, min_ram=None):
        if min_ram:
            return self._findFlavorByRam(min_ram, flavor_name)
        else:
            return self._findFlavorByName(flavor_name)

    # This method is wrapped with an LRU cache in the constructor.
    def _findNetwork(self, name):
        with Timer(self.log, 'API call get_network'):
            network = self._client.get_network(name)
            self.log.debug('Found network %s: %s', name, network)
        if not network:
            raise Exception("Unable to find network %s in endpoint %s" % (
                name, self.name))
        return network

    # This method is based on code from OpenStackSDK, licensed
    # under ASL2.
    def _simpleServerList(self):
        session = self._client.compute
        limit = None
        query_params = {}
        uri = '/servers/detail'
        ret = []
        while uri:
            response = session.get(
                uri,
                headers={"Accept": "application/json"},
                params=query_params,
            )
            data = response.json()

            last_marker = query_params.pop('marker', None)
            query_params.pop('limit', None)

            resources = data['servers']
            if not isinstance(resources, list):
                resources = [resources]

            ret += [ZuulOpenstackServer(x) for x in resources]

            if resources:
                marker = resources[-1]['id']
                uri, next_params = self._getNextLink(
                    uri, response, data, marker, limit
                )
                try:
                    if next_params['marker'] == last_marker:
                        raise Exception(
                            'Endless pagination loop detected, aborting'
                        )
                except KeyError:
                    pass
                query_params.update(next_params)
            else:
                break
        return ret

    # This method is based on code from OpenStackSDK, licensed
    # under ASL2.
    def _getNextLink(self, uri, response, data, marker, limit):
        pagination_key = 'servers_links'
        next_link = None
        params = {}

        if isinstance(data, dict):
            links = data.get(pagination_key, {})

            for item in links:
                if item.get('rel') == 'next' and 'href' in item:
                    next_link = item['href']
                    break

            if next_link and next_link.startswith('/v'):
                next_link = next_link[next_link.find('/', 1):]

        if not next_link and 'next' in response.links:
            # RFC5988 specifies Link headers and requests parses them if they
            # are there. We prefer link dicts in resource body, but if those
            # aren't there and Link headers are, use them.
            next_link = response.links['next']['uri']

        # Parse params from Link (next page URL) into params.
        # This prevents duplication of query parameters that with large
        # number of pages result in HTTP 414 error eventually.
        if next_link:
            parts = urllib.parse.urlparse(next_link)
            query_params = urllib.parse.parse_qs(parts.query)
            params.update(query_params)
            next_link = urllib.parse.urljoin(next_link, parts.path)

        # If we still have no link, and limit was given and is non-zero,
        # and the number of records yielded equals the limit, then the user
        # is playing pagination ball so we should go ahead and try once more.
        if not next_link and limit:
            next_link = uri
            params['marker'] = marker
            params['limit'] = limit

        return next_link, params

    def _listServers(self):
        with Timer(self.log, 'API call detailed server list'):
            return self._simpleServerList()

    def _listVolumes(self):
        try:
            with Timer(self.log, 'API call list_volumes'):
                return self._client.list_volumes()
        except EndpointNotFound:
            return []

    def _listFloatingIps(self):
        with Timer(self.log, 'API call list_floating_ips'):
            return self._client.list_floating_ips()

    def _refreshServer(self, obj):
        ret = self._getServer(obj['id'])
        if ret:
            return ret
        return obj

    def _expandServer(self, server):
        return openstack.cloud.meta.add_server_interfaces(
            self._client, server)

    def _getServer(self, external_id):
        for server in self._listServers():
            if server['id'] == external_id:
                if server['status'] in ['ACTIVE', 'ERROR']:
                    return self._expandServer(server)
                return server
        return None

    def _getServerByIdNow(self, server_id):
        # A synchronous get server by id.  Only to be used in error
        # handling where we can't wait for the list to update.
        with Timer(self.log, 'API call get_server_by_id'):
            return self._client.get_server_by_id(server_id)

    def _refreshServerDelete(self, obj):
        if obj is None:
            return obj
        for server in self._listServers():
            if server['id'] == obj['id']:
                if server['status'].lower() == 'deleted':
                    return None
                return server
        return None

    def _refreshFloatingIp(self, obj):
        for fip in self._listFloatingIps():
            if fip['id'] == obj['id']:
                return fip
        return obj

    def _refreshFloatingIpDelete(self, obj):
        if obj is None:
            return obj
        for fip in self._listFloatingIps():
            if fip['id'] == obj['id']:
                if fip.status == 'DOWN':
                    return None
                return fip
        return None

    def _needsFloatingIp(self, server):
        with Timer(self.log, 'API call _needs_floating_ip'):
            return self._client._needs_floating_ip(
                server=server, nat_destination=None)

    def _createFloatingIp(self, server):
        with Timer(self.log, 'API call create_floating_ip'):
            return self._client.create_floating_ip(server=server, wait=True)

    def _attachIpToServer(self, server, fip):
        # skip_attach is ignored for nova, which is the only time we
        # should actually call this method.
        with Timer(self.log, 'API call _attach_ip_to_server'):
            return self._client._attach_ip_to_server(
                server=server, floating_ip=fip,
                skip_attach=True)

    def _hasFloatingIps(self):
        # Not a network call
        return self._client._has_floating_ips()

    def _getFloatingIps(self, server):
        fips = openstack.cloud.meta.find_nova_interfaces(
            server['addresses'], ext_tag='floating')
        ret = []
        for fip in fips:
            with Timer(self.log, 'API call get_floating_ip'):
                ret.append(self._client.get_floating_ip(
                    id=None, filters={'floating_ip_address': fip['addr']}))
        return ret

    def _deleteFloatingIp(self, fip):
        with Timer(self.log, 'API call delete_floating_ip'):
            self._client.delete_floating_ip(fip['id'], retry=0)

    def _deleteServer(self, external_id):
        with Timer(self.log, 'API call delete_server'):
            self._client.delete_server(external_id)
        return True

    def _getFlavorFromServer(self, server):
        # In earlier versions of nova or the sdk, flavor has just an id.
        # In later versions it returns the information we're looking for.
        # If we get the information we want, we do not need to try to
        # lookup the flavor in our list.
        if hasattr(server['flavor'], 'vcpus'):
            return server['flavor']
        else:
            return self._findFlavorById(server['flavor']['id'])

    # The port cleanup logic.  We don't get tags or metadata, so we
    # have to figure this out on our own.

    # This method is not cached
    def _listPorts(self, status=None):
        '''
        List known ports.

        :param str status: A valid port status. E.g., 'ACTIVE' or 'DOWN'.
        '''
        if status:
            ports = self._client.list_ports(filters={'status': status})
        else:
            ports = self._client.list_ports()
        return ports

    def _filterComputePorts(self, ports):
        '''
        Return a list of compute ports (or no device owner).

        We are not interested in ports for routers or DHCP.
        '''
        ret = []
        for p in ports:
            if (p.device_owner is None or p.device_owner == '' or
                    p.device_owner.startswith("compute:")):
                ret.append(p)
        return ret

    def _cleanupLeakedPorts(self, interval):
        if not self._last_port_cleanup:
            self._last_port_cleanup = time.monotonic()
            ports = self._listPorts(status='DOWN')
            ports = self._filterComputePorts(ports)
            self._down_ports = set([p.id for p in ports])
            return

        # Return if not enough time has passed between cleanup
        last_check_in_secs = int(time.monotonic() - self._last_port_cleanup)
        if last_check_in_secs <= interval:
            return

        ports = self._listPorts(status='DOWN')
        ports = self._filterComputePorts(ports)
        current_set = set([p.id for p in ports])
        remove_set = current_set & self._down_ports

        removed_count = 0
        for port_id in remove_set:
            try:
                self._deletePort(port_id)
            except Exception:
                self.log.exception("Exception deleting port %s in %s:",
                                   port_id, self.name)
            else:
                removed_count += 1
                self.log.debug("Removed DOWN port %s in %s",
                               port_id, self.name)

        # if self._statsd and removed_count:
        #     key = 'nodepool.provider.%s.leaked.ports' % (self.name)
        #     self._statsd.incr(key, removed_count)

        self._last_port_cleanup = time.monotonic()

        # Rely on OpenStack to tell us the down ports rather than doing our
        # own set adjustment.
        ports = self._listPorts(status='DOWN')
        ports = self._filterComputePorts(ports)
        self._down_ports = set([p.id for p in ports])

    def _deletePort(self, port_id):
        self._client.delete_port(port_id)

    # This method is vendored from openstacksdk (Apache 2 licensed) then
    # modified to use our client object rather than self in order to add
    # additional debugging to the floating ip deletion process. This code
    # originates from:
    # https://opendev.org/openstack/openstacksdk/src/tag/4.8.0/openstack/cloud/_network_common.py#L971-L1001
    def _delete_unattached_floating_ips(self, retry=1):
        """Safely delete unattached floating ips.

        If the cloud can safely purge any unattached floating ips without
        race conditions, do so.

        Safely here means a specific thing. It means that you are not running
        this while another process that might do a two step create/attach
        is running. You can safely run this  method while another process
        is creating servers and attaching floating IPs to them if either that
        process is using add_auto_ip from shade, or is creating the floating
        IPs by passing in a server to the create_floating_ip call.

        :param retry: number of times to retry. Optional, defaults to 1,
            which is in addition to the initial delete call.
            A value of 0 will also cause no checking of results to occur.

        :returns: Number of Floating IPs deleted, False if none
        :raises: :class:`~openstack.exceptions.SDKException` on operation
            error.
        """
        processed = []
        if self._client._use_neutron_floating():
            for ip in self._client.list_floating_ips():
                if not bool(ip.port_id):
                    self.log.debug("Deleting unattached floating ip %s %s",
                                   ip['id'], ip['floating_ip_address'])
                    processed.append(
                        self._client.delete_floating_ip(
                            floating_ip_id=ip['id'], retry=retry
                        )
                    )
        return len(processed) if all(processed) else False

    def _cleanupFloatingIps(self):
        did_clean = self._delete_unattached_floating_ips()
        if did_clean:
            # some openstacksdk's return True if any port was
            # cleaned, rather than the count.  Just set it to 1 to
            # indicate something happened.
            if type(did_clean) is bool:
                did_clean = 1
            # if self._statsd:
            #     key = ('nodepool.provider.%s.leaked.floatingips'
            #            % self.name)
            #     self._statsd.incr(key, did_clean)

    def getConsoleLog(self, label, external_id):
        if not label.console_log:
            return None
        try:
            return self._client.get_server_console(external_id)
        except openstack.exceptions.OpenStackCloudException:
            return None



# Copyright 2021 Acme Gating, LLC
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from zuul.lib import strings

from tests.base import BaseTestCase


class TestStrings(BaseTestCase):

    def test_unique_project_name(self):
        self.assertEqual(('project', 'project'),
                         strings.unique_project_name('project'))
        self.assertEqual(('project', 'project%2Fsubproject'),
                         strings.unique_project_name('project/subproject'))
        self.assertEqual(('project', 'project%2Fsub%2Fproject'),
                         strings.unique_project_name('project/sub/project'))
